# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import random
from typing import List, Optional, TYPE_CHECKING, Type, Any, Tuple, Union, cast

from twisted.spread import pb
from twisted.python.failure import Failure
from twisted.internet.defer import Deferred, succeed, DeferredList
from qt import Qt, QTimer
from qt import QDialog, QVBoxLayout, QGridLayout, \
    QLabel, QPushButton, QWidget, \
    QProgressBar, QRadioButton, QSpacerItem, QSizePolicy

from kde import KIcon, KDialog
from dialogs import Sorry, Information, QuestionYesNo, KDialogIgnoringEscape
from guiutil import decorateWindow
from log import i18n, logWarning, logException, logDebug, logError, logFailure
from message import Message, ChatMessage
from chat import ChatWindow
from common import Options, SingleshotOptions, Internal, Debug, isAlive
from query import Query
from board import Board
from client import Client, ClientTable
from tables import TableList, SelectRuleset
from sound import Voice
from login import Connection
from rule import Ruleset
from player import PlayingPlayer
from game import PlayingGame
from visible import VisiblePlayingGame
from tile import Tile



if TYPE_CHECKING:
    from qt import QEvent, QKeyEvent
    from deferredutil import Request
    from move import Move
    from tile import Meld, MeldList
    from uitile import UITile
    from message import ClientMessage, ServerMessage
    from scene import PlayingScene
    from wind import Wind


class SelectChow(KDialogIgnoringEscape):

    """asks which of the possible chows is wanted"""

    def __init__(self, chows:'MeldList', propose:Optional['Meld'], deferred:Deferred) ->None:
        super().__init__()
        decorateWindow(self)
        self.setWindowFlags(cast(Qt.WindowType, Qt.WindowType.CustomizeWindowHint | Qt.WindowType.WindowTitleHint))
        self.setButtons(KDialog.NoButton)
        self.chows = chows
        self.selectedChow = None
        self.deferred = deferred
        layout = QVBoxLayout()
        label = QLabel(i18n('Which chow do you want to expose?'))
        layout.addWidget(label)
        layout.setAlignment(label, Qt.AlignmentFlag.AlignHCenter)
        self.buttons = []
        for chow in chows:
            button = QRadioButton('{}-{}-{}'.format(*(x.value for x in chow))) # pylint:disable=consider-using-f-string
            self.buttons.append(button)
            layout.addWidget(button)
            layout.setAlignment(button, Qt.AlignmentFlag.AlignHCenter)
            button.toggled.connect(self.toggled)
        widget = QWidget(self)
        widget.setLayout(layout)
        self.setMainWidget(widget)
        for idx, chow in enumerate(chows):
            if chow == propose:
                self.buttons[idx].setFocus()

    def toggled(self, unusedChecked:bool) ->None:
        """a radiobutton has been toggled"""
        button = cast(QRadioButton, self.sender())
        if button.isChecked():
            self.selectedChow = self.chows[self.buttons.index(button)]
            self.accept()
            self.deferred.callback((Message.Chow, self.selectedChow))


class SelectKong(KDialogIgnoringEscape):

    """asks which of the possible kongs is wanted"""

    def __init__(self, kongs:'MeldList', deferred:Deferred) ->None:
        super().__init__()
        decorateWindow(self)
        self.setButtons(KDialog.NoButton)
        self.kongs = kongs
        self.selectedKong = None
        self.deferred = deferred
        layout = QVBoxLayout()
        label = QLabel(i18n('Which kong do you want to declare?'))
        layout.addWidget(label)
        layout.setAlignment(label, Qt.AlignmentFlag.AlignHCenter)
        self.buttons = []
        for kong in kongs:
            button = QRadioButton((kong[0].name()), self)
            self.buttons.append(button)
            layout.addWidget(button)
            button.toggled.connect(self.toggled)
        widget = QWidget(self)
        widget.setLayout(layout)
        self.setMainWidget(widget)

    def toggled(self, unusedChecked:bool) ->None:
        """a radiobutton has been toggled"""
        button = cast(QRadioButton, self.sender())
        if button.isChecked():
            self.selectedKong = self.kongs[self.buttons.index(button)]
            self.accept()
            self.deferred.callback((Message.Kong, self.selectedKong))


class DlgButton(QPushButton):

    """special button for ClientDialog"""

    def __init__(self, message:'ClientMessage', parent:'ClientDialog') ->None:
        super().__init__(parent)
        self.message = message
        self.client:'HumanClient' = parent.client
        self.setMinimumHeight(25)
        self.setText(message.buttonCaption())  # type: ignore[call-arg]

    def setMeaning(self, uiTile:'UITile') ->None:
        """give me caption, shortcut, tooltip, icon"""
        txt, warn, _ = self.message.toolTip(self, uiTile.tile)
        if not txt:
            txt = self.message.i18nName  # .replace(i18nShortcut, '&'+i18nShortcut, 1)
        self.setToolTip(txt)
        self.setWarning(warn)

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """forward horizintal arrows to the hand board"""
        if event:
            key = Board.mapChar2Arrow(event)
            if key in [Qt.Key.Key_Left, Qt.Key.Key_Right]:
                game = self.client.game
                if game and game.activePlayer == game.myself:
                    if game.myself.handBoard:
                        game.myself.handBoard.keyPressEvent(event)
                    self.setFocus()
                    return
            super().keyPressEvent(event)

    def setWarning(self, warn:bool) ->None:
        """if warn, show a warning icon on the button"""
        if warn:
            self.setIcon(KIcon('dialog-warning'))
        else:
            self.setIcon(KIcon())


class ClientDialog(QDialog):  # pylint:disable=too-many-instance-attributes

    """a simple popup dialog for asking the player what he wants to do"""

    def __init__(self, client:'HumanClient', parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        decorateWindow(self, i18n('Choose'))
        self.tables:List[ClientTable]
        self.setObjectName('ClientDialog')
        self.client:'HumanClient' = client
        self.gridLayout = QGridLayout(self)
        self.progressBar = QProgressBar()
        self.progressBar.setMinimumHeight(25)
        self.timer = QTimer()
        assert client.game
        if not client.game.autoPlay:
            self.timer.timeout.connect(self.timeout)
        self.deferred:Optional[Deferred] = None
        self.buttons:List[DlgButton] = []
        self.setWindowFlags(cast(Qt.WindowType, Qt.WindowType.SubWindow | Qt.WindowType.WindowStaysOnTopHint))
        self.setModal(False)
        self.btnHeight = 0
        self.answered = False
        self.move:'Move'  # type:ignore[assignment]
        self.sorry:Optional[Sorry] = None

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """ESC selects default answer"""
        if not event or not self.client.game or self.client.game.autoPlay:
            return
        if event.key() in [Qt.Key.Key_Escape, Qt.Key.Key_Space]:
            self.selectButton()
            event.accept()
        else:
            for btn in self.buttons:
                if str(event.text()).upper() == btn.message.shortcut:
                    self.selectButton(btn.message)
                    event.accept()
                    return
            super().keyPressEvent(event)

    def __declareButton(self, message:'ClientMessage') ->None:
        """define a button"""
        if not self.client.game:
            return
        maySay = cast(PlayingPlayer, self.client.game.myself).sayable[message]
        assert Internal.Preferences
        if Internal.Preferences.showOnlyPossibleActions and not maySay:
            return
        btn = DlgButton(message, self)
        btn.setAutoDefault(True)
        btn.clicked.connect(self.selectedAnswer)
        self.buttons.append(btn)

    def focusTileChanged(self) ->None:
        """update icon and tooltip for the discard button"""
        if not self.client.game:
            return
        assert self.client.game.myself  # FIXME: needed?
        assert self.client.game.myself.handBoard
        newFocusTile = self.client.game.myself.handBoard.focusTile
        if newFocusTile:
            for button in self.buttons:
                button.setMeaning(newFocusTile)
        for uiTile in self.client.game.myself.handBoard.lowerHalfTiles():
            txt = []
            for button in self.buttons:
                _, _, tileTxt = button.message.toolTip(button, uiTile.tile)
                if tileTxt:
                    txt.append(tileTxt)
            uiTile.setToolTip(f'<font color=yellow>{"<br><br>".join(txt)}')
        if self.client.game.activePlayer == self.client.game.myself:
            if Internal.scene:
                Internal.scene.handSelectorChanged(
                    self.client.game.myself.handBoard)

    def checkTiles(self) ->None:
        """does the logical state match the displayed tiles?"""
        if not self.client.game:
            return
        for player in self.client.game.players:
            player.handBoard.checkTiles()

    def messages(self) ->List[Message]:
        """a list of all messages returned by the declared buttons"""
        return [x.message for x in self.buttons]

    def proposeAction(self) ->DlgButton:
        """either intelligently or first button by default. May also
        focus a proposed tile depending on the action."""
        result = self.buttons[0]
        game = self.client.game
        assert game
        assert game.myself
        assert game.myself.handBoard
        assert Internal.Preferences
        if game.autoPlay or Internal.Preferences.propose:
            mess = self.messages()
            answer, parameter = game.myself.intelligence.selectAnswer(mess)
            result = [x for x in self.buttons if x.message == answer][0]
            result.setFocus()
            if answer in [Message.Discard, Message.OriginalCall]:
                handBoard = game.myself.handBoard
                assert isinstance(parameter, Tile)
                handBoard.focusTile = handBoard[parameter][0]  # type:ignore[index]
        return result

    def askHuman(self, move:'Move', answers:List['ClientMessage'], deferred:Deferred) ->None:
        """make buttons specified by answers visible. The first answer is default.
        The default button only appears with blue border when this dialog has
        focus but we always want it to be recognizable. Hence setBackgroundRole."""
        self.move = move
        self.deferred = deferred
        for answer in answers:
            self.__declareButton(answer)
        self.focusTileChanged()
        self.show()
        self.checkTiles()
        game = self.client.game
        assert game
        myTurn = game.activePlayer == game.myself
        prefButton = self.proposeAction()
        if game.autoPlay:
            self.selectButton(prefButton.message)
            return
        prefButton.setFocus()

        self.progressBar.setVisible(not myTurn)
        if not myTurn:
            msecs = 50
            self.progressBar.setMinimum(0)
            self.progressBar.setMaximum(
                game.ruleset.claimTimeout * 1000 // msecs)
            self.progressBar.reset()
            self.timer.start(msecs)

    def placeInField(self) ->None:
        """place the dialog at bottom or to the right depending on space."""
        if not Internal.scene:
            logError('placeInField: have no Internal.scene')
            return
        mainWindow = Internal.scene.mainWindow
        cwi = mainWindow.centralWidget()
        assert cwi
        view = mainWindow.centralView
        geometry = self.geometry()
        if not self.btnHeight:
            self.btnHeight = self.buttons[0].height()
        vertical = view.width() > view.height() * 1.2
        if vertical:
            height = (len(self.buttons) + 1) * self.btnHeight * 1.2
            width = (cwi.width() - cwi.height()) // 2
            geometry.setX(int(cwi.width() - width))
            geometry.setY(int(min(cwi.height() // 3, cwi.height() - height)))
        else:
            assert self.client.game
            handBoard = self.client.game.myself.handBoard
            if not handBoard:
                # we are in the progress of logging out
                return
            hbLeftTop = view.mapFromScene(
                handBoard.mapToScene(handBoard.rect().topLeft()))
            hbRightBottom = view.mapFromScene(
                handBoard.mapToScene(handBoard.rect().bottomRight()))
            width = hbRightBottom.x() - hbLeftTop.x()
            height = self.btnHeight
            geometry.setY(cwi.height() - height)
            geometry.setX(hbLeftTop.x())
        for idx, btn in enumerate(self.buttons + [self.progressBar]):
            self.gridLayout.addWidget(
                btn,
                idx +
                1 if vertical else 0,
                idx +
                1 if not vertical else 0)
        idx = len(self.buttons) + 2
        spacer = QSpacerItem(
            20,
            20,
            QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(
            spacer,
            idx if vertical else 0,
            idx if not vertical else 0)

        geometry.setWidth(int(width))
        geometry.setHeight(int(height))
        self.setGeometry(geometry)

    def showEvent(self, unusedEvent:Optional['QEvent']) ->None:
        """try to place the dialog such that it does not cover interesting information"""
        self.placeInField()

    def timeout(self) ->None:
        """the progressboard wants an update"""
        pBar = self.progressBar
        if isAlive(pBar):
            pBar.setValue(pBar.value() + 1)
            pBar.setVisible(True)
            if pBar.value() == pBar.maximum():
                # timeout: we always return the original default answer, not
                # the one with focus
                self.selectButton()
                pBar.setVisible(False)

    def selectButton(self, message:Optional[Message]=None) ->None:
        """select default answer. button may also be of type Message."""
        if self.answered:
            # sometimes we get this event twice
            return
        if message is None:
            message = cast(DlgButton, self.focusWidget()).message
        assert any(x.message == message for x in self.buttons)
        assert self.client.game
        if not cast(PlayingPlayer, self.client.game.myself).sayable[message]:
            self.proposeAction().setFocus() # go back to default action
            self.sorry = Sorry(i18n('You cannot say %1', message.i18nName))
            return
        self.timer.stop()
        self.answered = True
        if self.sorry:
            self.sorry.cancel()
        self.sorry = None
        if Internal.scene:
            Internal.scene.clientDialog = None
        assert self.deferred
        self.deferred.callback(message)

    def selectedAnswer(self, unusedChecked:bool) ->None:
        """the user clicked one of the buttons"""
        game = self.client.game
        if game and not game.autoPlay:
            self.selectButton(cast(DlgButton, self.sender()).message)


class HumanClient(Client):

    """a human client"""
    humanClients : List['HumanClient'] = []

    def __init__(self) ->None:
        super().__init__()
        HumanClient.humanClients.append(self)
        self.table = None
        self.ruleset:Ruleset
        self.connection:Optional[Connection]
        self.beginQuestion:Optional[Deferred] = None
        self.tableList:Optional[TableList] = TableList(self)
        Connection(self).login().addCallbacks(
            self.__loggedIn,
            self.__loginFailed)

    @staticmethod
    def shutdownHumanClients(exception:Optional[Exception]=None) ->Any: # Union[Deferred, DeferredList]:
        """close connections to servers except maybe one"""
        clients = HumanClient.humanClients

        def done() ->bool:
            """return True if clients is cleaned"""
            return len(clients) == 0 or (exception is not None and clients == [exception])

        def disconnectedClient(unusedResult:Deferred, client:'HumanClient') ->None:
            """now the client is really disconnected from the server"""
            if client in clients:
                # HumanClient.serverDisconnects also removes it!
                clients.remove(client)
        if isinstance(exception, Failure):
            logException(exception)
        for client in clients[:]:
            if client.tableList:
                client.tableList.hide()
        if done():
            return succeed(None)
        deferreds = []
        for client in clients[:]:
            if client != exception and client.connection:
                deferreds.append(
                    client.logout(
                    ).addCallback(
                        disconnectedClient,
                        client).addErrback(logFailure))
        return DeferredList(deferreds)

    def __loggedIn(self, connection:Connection) ->None:
        """callback after the server answered our login request"""
        self.connection = connection
        self.ruleset = connection.ruleset
        self.name = connection.username
        if self.tableList:
            self.tableList.show()
        voiceId = None
        assert Internal.Preferences
        if Internal.Preferences.uploadVoice:
            voice = Voice.locate(self.name)
            if voice:
                voiceId = voice.md5sum
            if Debug.sound and voiceId:
                logDebug(
                    f'{self.name} sends own voice {voiceId} to server')
        maxGameId = Query('select max(id) from game').records[0][0]
        maxGameId = int(maxGameId) if maxGameId else 0
        self.callServer('setClientProperties',
                        Internal.db.identifier,
                        voiceId, maxGameId,
                        Internal.defaultPort).addCallbacks(self.__initTableList, self.__versionError)

    def __initTableList(self, unused:str) ->None:
        """first load of the list. Process options like --demo, --table, --join"""
        self.showTableList()
        if SingleshotOptions.table:
            Internal.autoPlay = False
            self.__requestNewTableFromServer(SingleshotOptions.table).addCallback(
                self.__showTables).addErrback(self.tableError)
            if Debug.table:
                logDebug(
                    f'{self.name}: --table lets us open a new table')
            SingleshotOptions.table = False
        elif SingleshotOptions.join:
            Internal.autoPlay = False
            self.callServer('joinTable', SingleshotOptions.join).addCallback(
                self.__showTables).addErrback(self.tableError)
            if Debug.table:
                logDebug(
                    f'{self.name}: --join lets us join table {self._tableById(SingleshotOptions.join)}')
            SingleshotOptions.join = False
        elif not self.game and (Internal.autoPlay or (not self.tables and self.hasLocalServer())):
            self.__requestNewTableFromServer().addCallback(
                self.__newLocalTable).addErrback(self.tableError)
        else:
            self.__showTables()

    @staticmethod
    def __loginFailed(unused:List[Deferred]) ->None:
        """as the name says"""
        if Internal.scene:
            cast('PlayingScene', Internal.scene).startingGame = False

    def isRobotClient(self) ->bool:
        """avoid using isinstance, it would import too much for kajonggserver"""
        return False

    @staticmethod
    def isHumanClient() ->bool:
        """avoid using isinstance, it would import too much for kajonggserver"""
        return True

    def isServerClient(self) ->bool:
        """avoid using isinstance, it would import too much for kajonggserver"""
        return False

    def hasLocalServer(self) ->bool:
        """True if we are talking to a Local Game Server"""
        return self.connection is not None and self.connection.url.isLocalHost

    def __updateTableList(self) ->None:
        """if it exists"""
        if self.tableList:
            self.tableList.loadTables(self.tables)

    def __showTables(self, unused:Optional[str]=None) ->None:
        """load and show tables. We may be used as a callback. In that case,
        clientTables is the id of a new table - which we do not need here"""
        if self.tableList:
            self.tableList.loadTables(self.tables)
            self.tableList.show()

    def showTableList(self, unused:Optional[str]=None) ->None:
        """allocate it if needed"""
        if not self.tableList:
            self.tableList = TableList(self)
        self.tableList.loadTables(self.tables)
        self.tableList.activateWindow()

    def remote_tableRemoved(self, tableid:int, message:str, *args:Any) ->None:
        """update table list"""
        super().remote_tableRemoved(tableid, message, *args)
        self.__updateTableList()
        if message:
            if self.name not in args or not message.endswith('has logged out'):
                logWarning(i18n(message, *args))

    def __receiveTables(self, tables:List[List[Any]]) ->None:
        """now we already know all rulesets for those tables"""
        super().remote_newTables(tables)
        if not Internal.autoPlay:
            if self.hasLocalServer():
                # when playing a local game, only show pending tables with
                # previously selected ruleset
                self.tables = [x for x in self.tables if x.ruleset == self.ruleset]
        if self.tables:
            self.__updateTableList()

    def remote_newTables(self, tables:List[List[Any]]) ->None:
        """update table list"""
        assert tables

        def gotRulesets(result:List[Union[int, str]]) ->List[List[Any]]:
            """the server sent us the wanted ruleset definitions"""
            for ruleset in result:
                Ruleset.cached(ruleset).save()  # make it known to the cache and save in db
            return tables
        rulesetHashes = {x[1] for x in tables}
        needRulesets = [x for x in rulesetHashes if not Ruleset.hashIsKnown(x)]
        if needRulesets:
            _ = self.callServer('needRulesets', needRulesets)
            _ = _.addCallback(gotRulesets)
            _ = _.addErrback(logFailure)
            _ = _.addCallback(self.__receiveTables)
            _.addErrback(logFailure)
        else:
            self.__receiveTables(tables)

    @staticmethod
    def remote_needRuleset(ruleset:str) ->List[List[Union[str, int, float]]]:
        """server only knows hash, needs full definition"""
        result = Ruleset.cached(ruleset)
        assert result and result.hash == ruleset
        return result.toList()

    def tableChanged(self, table:ClientTable) ->Tuple[Optional[ClientTable], ClientTable]:
        """update table list"""
        oldTable, newTable = super().tableChanged(table)
        if oldTable and oldTable == self.table:
            # this happens if a table has more than one human player and
            # one of them leaves the table. In that case, the other players
            # need this code.
            self.table = newTable
            if len(newTable.playerNames) == 3:
                # only tell about the first player leaving, because the
                # others will then automatically leave too
                for name in oldTable.playerNames:
                    if name != self.name and not newTable.isOnline(name):
                        def sorried(unused:List['Request']) ->Optional[Deferred]:
                            """user ack"""
                            game = self.game
                            if game:
                                self.game = None
                                return game.close()
                            return None
                        if self.beginQuestion:
                            self.beginQuestion.cancel()
                        Sorry(i18n('Player %1 has left the table', name)).addCallback(
                            sorried).addErrback(logFailure).addCallback(self.showTableList).addErrback(logFailure)
                        break
        self.__updateTableList()
        return oldTable, newTable

    def remote_chat(self, data):
        """others chat to me"""
        chatLine = ChatMessage(data)
        if Debug.chat:
            logDebug(f'got chatLine: {chatLine}')
        table = self._tableById(chatLine.tableid)
        if not chatLine.isStatusMessage and not table.chatWindow:
            ChatWindow(table=table)
        if table.chatWindow:
            table.chatWindow.receiveLine(chatLine)

    def readyForGameStart(
            self, tableid:int, gameid:int, wantedGame:str, playerNames:List[Tuple['Wind', str]], shouldSave:bool=True,
            gameClass:Optional[Type]=None) ->Deferred:
        """playerNames are in wind order ESWN"""
        if gameClass is None:
            if Options.gui:
                gameClass = VisiblePlayingGame
            else:
                gameClass = PlayingGame

        def clientReady() ->Deferred:
            """macro"""
            return Client.readyForGameStart(
                self, tableid, gameid, wantedGame, playerNames,
                shouldSave, gameClass)

        def answered(result:Deferred) ->Union[Message, Deferred]:
            """callback, called after the client player said yes or no"""
            self.beginQuestion = None
            if self.connection and result:
                # still connected and yes, we are
                return clientReady()
            return Message.NoGameStart

        def cancelled(unused:Failure) ->'ServerMessage':
            """the user does not want to start now. Back to table list"""
            if Debug.table:
                logDebug(f'{self.name}: Readyforgamestart returns Message.NoGameStart '
                         f'for table {self._tableById(tableid)}')
            self.table = None
            self.beginQuestion = None
            if self.tableList:
                self.__updateTableList()
                self.tableList.show()
            return Message.NoGameStart
        if sum(not x[1].startswith('Robot ') for x in playerNames) == 1:
            # we play against 3 robots and we already told the server to start:
            # no need to ask again
            return clientReady()
        assert not self.table
        assert self.tables
        self.table = self._tableById(tableid)
        if not self.table:
            raise pb.Error(
                f'client.readyForGameStart: tableid {int(tableid)} unknown')
        msg = i18n(
            "The game on table <numid>%1</numid> can begin. Are you ready to play now?",
            tableid)
        self.beginQuestion = QuestionYesNo(msg, modal=False, caption=self.name).addCallback(
            answered).addErrback(cancelled)
        return self.beginQuestion

    def readyForHandStart(self, playerNames:List[Tuple['Wind', str]], mustRotateWinds:bool) ->Optional[Deferred]:
        """playerNames are in wind order ESWN. Never called for first hand."""
        def answered(unused:Optional[List['Request']]=None) ->Optional[Deferred]:
            """called after the client player said yes, I am ready"""
            return Client.readyForHandStart(self, playerNames, mustRotateWinds) if self.connection else None
        if not self.connection:
            # disconnected meanwhile
            return None
        if Options.gui:
            # update the balances in the status bar:
            if Internal.mainWindow:
                Internal.mainWindow.updateGUI()
        assert self.game
        assert not self.game.isFirstHand()
        return Information(i18n("Ready for next hand?"), modal=False).addCallback(answered).addErrback(logFailure)

    def ask(self, move:'Move', answers:List['ClientMessage']) ->Deferred:
        """server sends move. We ask the user. answers is a list with possible answers,
        the default answer being the first in the list."""
        if not Options.gui:
            return super().ask(move, answers)
        scene = Internal.scene
        assert scene
        assert self.game
        cast(PlayingPlayer, self.game.myself).computeSayable(move, answers)
        deferred:Deferred = Deferred()
        deferred.addCallback(self.__askAnswered)
        deferred.addErrback(self.__answerError, move, answers)
        iAmActive = self.game.myself == self.game.activePlayer
        assert self.game.myself.handBoard
        self.game.myself.handBoard.setEnabled(iAmActive)
        oldDialog = scene.clientDialog
        assert oldDialog is None or oldDialog.answered, (
            f'old dialog {str(oldDialog.move)}:{str([x.message.name for x in oldDialog.buttons])} '
            f'is unanswered, new Dialog: {str(move)}/{str(answers)}')
        if not oldDialog or not oldDialog.isVisible():
            # always build a new dialog because if we change its layout before
            # reshowing it, sometimes the old buttons are still visible in which
            # case the next dialog will appear at a lower position than it
            # should
            scene.clientDialog = ClientDialog(
                self,
                scene.mainWindow.centralWidget())
        assert scene.clientDialog
        assert scene.clientDialog.client is self
        scene.clientDialog.askHuman(move, answers, deferred)
        return deferred

    def __selectChow(self, chows:'MeldList') ->Union[Deferred, Tuple['ClientMessage', Union['Meld', None]]]:
        """which possible chow do we want to expose?
        Since we might return a Deferred to be sent to the server,
        which contains Message.Chow plus selected Chow, we should
        return the same tuple here"""
        assert self.game
        intelligence = self.game.myself.intelligence
        if self.game.autoPlay:
            return cast('ClientMessage', Message.Chow), intelligence.selectChow(chows)
        if len(chows) == 1:
            return cast('ClientMessage', Message.Chow), chows[0]
        assert Internal.Preferences
        if Internal.Preferences.propose:
            propose = intelligence.selectChow(chows)
        else:
            propose = None
        deferred:Deferred = Deferred()
        selDlg = SelectChow(chows, propose, deferred)
        assert selDlg.exec()
        return deferred

    def __selectKong(self, kongs:'MeldList') ->Union[Deferred, Tuple['ClientMessage', Optional['Meld']]]:
        """which possible kong do we want to declare?"""
        assert self.game
        if self.game.autoPlay:
            return cast('ClientMessage', Message.Kong), self.game.myself.intelligence.selectKong(kongs)
        if len(kongs) == 1:
            return cast('ClientMessage', Message.Kong), kongs[0]
        deferred:Deferred = Deferred()
        selDlg = SelectKong(kongs, deferred)
        assert selDlg.exec()
        return deferred

    def __askAnswered(self, answer:'ClientMessage') ->Union[
            Deferred,
            Tuple['ClientMessage', Union[bool, 'Tile', 'Meld',
                'MeldList', Tuple['MeldList', Optional['Tile'], 'Meld'], None ]]]:
        """the user answered our question concerning move"""
        if not self.game:
            return cast('ClientMessage', Message.NoClaim), None
        myself = self.game.myself
        if answer in [Message.Discard, Message.OriginalCall]:
            # do not remove tile from hand here, the server will tell all players
            # including us that it has been discarded. Only then we will remove
            # it.
            assert myself.handBoard
            assert myself.handBoard.focusTile
            myself.handBoard.setEnabled(False)
            return answer, myself.handBoard.focusTile.tile
        args = cast(PlayingPlayer, myself).sayable[answer]
        assert args
        if answer == Message.Chow:
            return self.__selectChow(cast('MeldList', args))
        if answer == Message.Kong:
            return self.__selectKong(cast('MeldList', args))
        self.game.hidePopups()
        if args is True or args == []:
            # this does not specify any tiles, the server does not need this. Robot players
            # also return None in this case.
            return answer, None
        return answer, args


    def __answerError(self, result: Any, move:'Move', answers:List['ClientMessage']) ->None:
        """an error happened while determining the answer to server"""
        logException(
            f"{result} {self.game.myself.name if self.game else 'NOGAME'} {move} {answers}")

    def remote_abort(self, tableid:int, message: str, *args:Any) ->None:
        """the server aborted this game"""
        if self.table and self.table.tableid == tableid:
            # translate Robot to Roboter:
            if self.game:
                args = self.game.players.translatePlayerNames(args)  # type:ignore[assignment]
            logWarning(i18n(message, *args))
            if self.game:
                self.game.close()
                self.game = None

    def remote_gameOver(self, tableid:int, message:str, *args:Any) ->None:
        """the game is over"""
        def yes(unused:'Request') ->None:
            """now that the user clicked the 'game over' prompt away, clean up"""
            if self.game:
                self.game.rotateWinds()
                if self.game.finished():
                    self.game.finish_in_db()
                _ = self.game.close()
                if Internal.mainWindow:
                    _.addCallback(Internal.mainWindow.close).addErrback(logFailure)
        assert self.table and self.table.tableid == tableid
        if Internal.scene:
            # update the balances in the status bar:
            Internal.scene.mainWindow.updateGUI()
        Information(i18n(message, *args)).addCallback(yes).addErrback(logFailure)

    def remote_serverDisconnects(self, result:Optional[str]=None) ->None:
        """we logged out or lost connection to the server.
        Remove visual traces depending on that connection."""
        if Debug.connections and result:
            logDebug(
                f'server {self.connection} disconnects: {result}')
        self.connection = None
        game = self.game
        self.game = None  # avoid races: messages might still arrive
        if self.tableList:
            self.tableList.hide()
            self.tableList = None
        if self in HumanClient.humanClients:
            HumanClient.humanClients.remove(self)
        if self.beginQuestion:
            self.beginQuestion.cancel()
        scene = Internal.scene
        if scene and game and scene.game == game:
            scene.game = None
        if not Options.gui:
            if Internal.mainWindow:
                Internal.mainWindow.close()

    def serverDisconnected(self, unusedReference:Deferred) ->None:
        """perspective calls us back"""
        if self.connection and (Debug.traffic or Debug.connections):
            logDebug(
                f'perspective notifies disconnect: {self.connection.url}')
        self.remote_serverDisconnects()

    @staticmethod
    def __versionError(err:Failure) ->Failure:
        """log the twisted error"""
        logWarning(err.getErrorMessage())
        if Internal.game:
            Internal.game.close()
            Internal.game = None
        return err

    @staticmethod
    def __wantedGame() ->str:
        """find out which game we want to start on the table"""
        result = SingleshotOptions.game
        if not result or result == '0':
            result = str(int(random.random() * 10 ** 9))
        SingleshotOptions.game = None
        return result

    def tableError(self, err:Failure) ->None:
        """log the twisted error"""
        if not self.connection:
            # lost connection to server
            if self.tableList:
                self.tableList.hide()
                self.tableList = None
        else:
            logWarning(err.getErrorMessage())

    def __newLocalTable(self, newId:int) ->Deferred:
        """we just got newId from the server"""
        return self.callServer('startGame', newId).addErrback(self.tableError)

    def __requestNewTableFromServer(self, tableid:Optional[int]=None, ruleset:Optional[Ruleset]=None) ->Deferred:
        """as the name says"""
        if ruleset is None:
            ruleset = self.ruleset
        assert self.connection
        self.connection.ruleset = ruleset  # side effect: saves ruleset as last used for server
        return self.callServer('newTable', ruleset.hash, Options.playOpen,
                               Internal.autoPlay, self.__wantedGame(), tableid).addErrback(self.tableError)

    def newTable(self) ->None:
        """TableList uses me as a slot"""
        if Options.ruleset:
            ruleset = Options.ruleset
        elif self.hasLocalServer():
            ruleset = self.ruleset
        else:
            assert self.connection
            selectDialog = SelectRuleset(self.connection.url)
            if not selectDialog.exec():
                return
            ruleset = selectDialog.cbRuleset.current
        deferred = self.__requestNewTableFromServer(ruleset=ruleset)
        if self.hasLocalServer():
            deferred.addCallback(self.__newLocalTable).addErrback(logFailure)
        assert self.tableList
        self.tableList.requestedNewTable = True

    def joinTable(self, table:Optional[ClientTable]=None) ->None:
        """join a table"""
        if not isinstance(table, ClientTable):
            assert self.tableList
            table = self.tableList.selectedTable()
            assert table
        self.callServer('joinTable', table.tableid).addErrback(self.tableError)

    def logout(self, unusedResult:Optional[List['Request']]=None) ->Deferred:
        """clean visual traces and logout from server"""
        def loggedout(result:List['Request'], connection:Connection) ->List['Request']:
            """end the connection from client side"""
            if Debug.connections:
                logDebug(f'server confirmed logout for {self}')
            connection.connector.disconnect()
            return result
        if self.connection:
            conn = self.connection
            self.connection = None
            if Debug.connections:
                logDebug(f'sending logout to server for {self}')
            return self.callServer('logout').addCallback(loggedout, conn).addErrback(logFailure)
        return succeed(None)

    def __logCallServer(self, *args:Any) ->None:
        """for Debug.traffic"""
        debugArgs = list(args[:])
        if Debug.neutral:
            if debugArgs[0] == 'ping':
                return
            if debugArgs[0] == 'setClientProperties':
                debugArgs[1] = 'DBID'
                debugArgs[3] = 'GAMEID'
                if debugArgs[4] >= 8300:
                    debugArgs[4] -= 300
        if self.game:
            self.game.debug(f'callServer({repr(debugArgs)})')
        else:
            logDebug(f'callServer({repr(debugArgs)})')

    def callServer(self, *args:Any) ->Deferred:
        """if we are online, call server"""
        if self.connection:
            if args[0] is None:
                args = args[1:]
            try:
                if Debug.traffic:
                    self.__logCallServer(*args)

                def callServerError(result:str) ->Optional[str]:
                    """if serverDisconnected has been called meanwhile, just ignore msg about
                    connection lost in a non-clean fashion"""
                    return result if self.connection else None
                return self.connection.perspective.callRemote(*args).addErrback(callServerError)
            except pb.DeadReferenceError:
                logWarning(
                    i18n(
                        'The connection to the server %1 broke, please try again later.',
                        self.connection.url))
                self.remote_serverDisconnects()
                return succeed(None)
        else:
            return succeed(None)

    def sendChat(self, chatLine:'ChatMessage') ->Deferred:
        """send chat message to server"""
        return self.callServer('chat', chatLine.asList())



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import weakref
from typing import Dict, List, Type, Optional, Set, Generator, Tuple, TYPE_CHECKING
from typing import Union, Any, Callable, overload, SupportsIndex, Iterable, cast

from log import logException, logWarning, logDebug
from mi18n import i18n, i18nc, i18nE
from common import IntDict, Debug
from common import ReprMixin, Internal
from wind import East, Wind
from query import Query
from tile import Tile, TileList, TileTuple, PieceList, elements, Meld, MeldList
from tilesource import TileSource
from permutations import Permutations
from message import Message
from hand import Hand
from intelligence import AIDefaultAI

if TYPE_CHECKING:
    from sound import Voice
    from tile import Tiles, Piece
    from game import Game, PlayingGame
    from uitile import UITile
    from deferredutil import Request
    from rule import UsedRule
    from move import Move
    from handboard import HandBoard
    from message import ClientMessage


class Players(list, ReprMixin):

    """a list of players where the player can also be indexed by wind.
    The position in the list defines the place on screen. First is on the
    screen bottom, second on the right, third top, forth left"""

    allNames:Dict[int, str] = {}
    allIds:Dict[str, int] = {}
    humanNames:Dict[int, str] = {}

    def __init__(self, players:Optional[List['Player']]=None) ->None:
        list.__init__(self)
        if players:
            self.extend(players)

    @overload
    def __getitem__(self, index:SupportsIndex) ->'Player': ...
    @overload
    def __getitem__(self, index:slice) ->List['Player']: ...

    def __getitem__(self, index:Union[SupportsIndex, slice]) ->Union['Player', List['Player']]:
        """allow access by idx or by wind"""
        if isinstance(index, Wind):
            for player in self:
                if player.wind == index:
                    return player
        assert isinstance(index, (int, slice)), f'index is neither Wind, int nor slice:{type(index)}'
        return super().__getitem__(index)

    def __str__(self) ->str:
        return ', '.join(f'{x.name}: {x.wind}' for x in self)

    def byId(self, playerid:int) ->Optional['Player']:
        """lookup the player by id"""
        for player in self:
            if player.nameid == playerid:
                return player
        logException(f"no player has id {int(playerid)}")
        return None

    def byName(self, playerName:str) ->Optional['Player']:
        """lookup the player by name"""
        for player in self:
            if player.name == playerName:
                return player
        logException(
            f"no player has name '{playerName}' - we have {[x.name for x in self]}")
        return None

    @staticmethod
    def load() -> None:
        """load all defined players into self.allIds and self.allNames"""
        Players.allIds = {}
        Players.allNames = {}
        for nameid, name in Query("select id,name from player").records:
            Players.allIds[name] = nameid
            Players.allNames[nameid] = name
            if not name.startswith('Robot'):
                Players.humanNames[nameid] = name

    @staticmethod
    def createIfUnknown(name:str) ->None:
        """create player in database if not there yet"""
        if not Internal.db:
            # kajonggtest
            nameid = len(Players.allIds) + 1
            Players.allIds[name] = nameid
            Players.allNames[nameid] = name
            if not name.startswith('Robot'):
                Players.humanNames[nameid] = name

        if name not in Players.allNames.values():
            Players.load()  # maybe somebody else already added it
            if name not in Players.allNames.values():
                Query("insert or ignore into player(name) values(?)", (name,))
                Players.load()
        assert name in Players.allNames.values(), f'{name} not in {Players.allNames.values()}'

    def translatePlayerNames(self, names:Iterable[str]) ->List[str]:
        """for a list of names, translates those names which are english
        player names into the local language"""
        known = {x.name for x in self}
        return [self.byName(x).localName if x in known else x for x in names]  # type: ignore[union-attr]


class Player(ReprMixin):

    """
    all player related attributes without GUI stuff.
    concealedTiles: used during the hand for all concealed tiles, ungrouped.
    concealedMelds: is empty during the hand, will be valid after end of hand,
    containing the concealed melds as the player presents them.

    @todo: Now that Player() always calls createIfUnknown, test defining new
    players and adding new players to server
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, game:Optional['Game'], name:str) ->None:
        """
        Initialize a player for a give game.

        @type game: L{Game} or None.
        @param game: The game this player is part of. May be None.
        """
        self._game:Optional[weakref.ReferenceType['Game']] = None
        if game:
            self._game = weakref.ref(game)
        self.__balance = 0
        self.__payment = 0
        self.wonCount = 0
        self.__name = ''
        Players.createIfUnknown(name)
        self.name = name
        self.wind:Wind = East
        self.intelligence:AIDefaultAI = AIDefaultAI(self)  # type:ignore[arg-type]
        self.visibleTiles:Dict[Tile, int] = IntDict(cast(IntDict, game.visibleTiles)) if game else IntDict()
        self.handCache:Dict[str, 'Hand'] = {}
        self.cacheHits:int = 0
        self.cacheMisses:int = 0
        self.__lastSource:Type[TileSource.SourceClass] = TileSource.Unknown
        self.clearHand()
        self.handBoard:Optional['HandBoard'] = None

    def __lt__(self, other:Any) ->bool:
        """Used for sorting"""
        if not other:
            return False
        if not isinstance(other, Player):
            return False
        return self.name < other.name

    def clearCache(self) -> None:
        """clears the cache with Hands"""
        if Debug.hand and self.handCache and self.game:
            self.game.debug(
                f'{self}: cache hits:{int(self.cacheHits)} misses:{int(self.cacheMisses)}')
        self.handCache.clear()
        Permutations.cache.clear()
        self.cacheHits = 0
        self.cacheMisses = 0

    @property
    def name(self) ->str:
        """
        The name of the player, can be changed only once.

        @type: C{str}
        """
        return self.__name

    @name.setter
    def name(self, value:str) ->None:
        """write once"""
        assert self.__name == ''
        assert value
        assert isinstance(value, str), f'Player.name must be str but not {type(value)}'
        self.__name = value

    @property
    def game(self) ->Optional['Game']:
        """hide the fact that this is a weakref"""
        return self._game() if self._game else None

    def clearHand(self) -> None:
        """clear player attributes concerning the current hand"""
        self._concealedTiles:PieceList = PieceList()
        self._exposedMelds = MeldList()
        self._concealedMelds = MeldList()
        self._bonusTiles:TileList = TileList()
        self.discarded:List[Tile] = []
        self.visibleTiles.clear()
        self.newHandContent = None
        self.originalCallingHand:Optional[Hand] = None
        self.__lastTile = Tile.none
        self.lastSource = TileSource.Unknown
        self.lastMeld = Meld()
        self.__mayWin = True
        self.__payment = 0
        self.__originalCall:bool = False
        self.dangerousTiles:List[Tuple[Set[Tile], str]] = []
        self.claimedNoChoice:bool = False
        self.playedDangerous:bool = False
        self.usedDangerousFrom:Optional['PlayingPlayer'] = None
        self.isCalling:bool = False
        self.clearCache()
        self._hand:Optional['Hand'] = None

    @property
    def lastTile(self) ->Tile:
        """temp for debugging"""
        return self.__lastTile

    @lastTile.setter
    def lastTile(self, value:Tile) ->None:
        """temp for debugging"""
        assert isinstance(value, Tile), value
        self.__lastTile = value
        if value is Tile.none:
            self.lastMeld = Meld()
        # FIXME self.assertLastTile()
# this would probably work if we compute new tiles first and lastTile/lastMeld last, but
# when changing that order, kajongg returns different results.

    @property
    def originalCall(self) ->bool:
        """temp for debugging"""
        return self.__originalCall

    @originalCall.setter
    def originalCall(self, value:bool) ->None:
        """temp for debugging"""
        if self.__originalCall != value:
            self.__originalCall = value
            self._hand = None  # force recreation

    def invalidateHand(self) ->None:
        """some source for the computation of current hand changed"""
        self._hand = None

    @property
    def hand(self) ->'Hand':
        """readonly: the current Hand. Compute if invalidated."""
        if not self._hand:
            self._hand = self.__computeHand()
        elif Debug.hand:
            _ = self.__computeHand()
            assert self._hand == _, f'{self.hand} != {_}'
        return self._hand

    @property
    def bonusTiles(self) ->TileList:
        """a readonly tuple"""
        return self._bonusTiles

    @property
    def concealedTiles(self) ->TileTuple:
        """a readonly tuple"""
        return TileTuple(self._concealedTiles)

    @property
    def exposedMelds(self) ->MeldList:
        """a readonly tuple"""
        return self._exposedMelds

    @property
    def concealedMelds(self) ->MeldList:
        """a readonly tuple"""
        return self._concealedMelds

    @property
    def mayWin(self) ->bool:
        """winning possible?"""
        return self.__mayWin

    @mayWin.setter
    def mayWin(self, value:bool) ->None:
        """winning possible?"""
        if self.__mayWin != value:
            self.__mayWin = value
            self._hand = None

    @property
    def lastSource(self) ->Type[TileSource.SourceClass]:
        """the source of the last tile the player got"""
        return self.__lastSource

    @lastSource.setter
    def lastSource(self, value: Type[TileSource.SourceClass]) ->None:
        """the source of the last tile the player got"""
        if self.game:
            if self.game.wall:
                if value is TileSource.LivingWallDiscard and not self.game.wall.living:
                    value = TileSource.LivingWallEndDiscard
                if value is TileSource.LivingWall and not self.game.wall.living:
                    value = TileSource.LivingWallEnd
        if self.__lastSource != value:
            self.__lastSource = value
            self._hand = None

    @property
    def nameid(self) ->int:
        """the name id of this player"""
        return Players.allIds[self.name]

    @property
    def localName(self) ->str:
        """the localized name of this player"""
        return i18nc('kajongg, name of robot player, to be translated', self.name)

    @property
    def handTotal(self) ->int:
        """the hand total of this player for the final scoring"""
        assert self.game
        return 0 if not self.game.winner else self.hand.total()

    @property
    def balance(self) ->int:
        """the balance of this player"""
        return self.__balance

    @balance.setter
    def balance(self, balance:int) ->None:
        """the balance of this player"""
        self.__balance = balance
        self.__payment = 0

    def getsPayment(self, payment:int) ->None:
        """make a payment to this player"""
        self.__balance += payment
        self.__payment += payment

    @property
    def payment(self) ->int:
        """the payments for the current hand"""
        return self.__payment

    @payment.setter
    def payment(self, payment:int) ->None:
        """the payments for the current hand"""
        assert payment == 0
        self.__payment = 0

    def __str__(self) ->str:
        return f'{self.name[:10]:<10} {self.wind}'

    def pickedTile(self, deadEnd:bool, tileName:Optional[Tile] =None) -> Tile:
        """got a tile from wall"""
        assert self.game
        assert self.game.wall
        self.game.activePlayer = self
        tile = self.game.wall.deal(tileName, deadEnd=deadEnd)[0]  # type:ignore[list-item]
        if hasattr(tile, 'tile'):
            self.lastTile = tile.tile
        else:
            self.lastTile = tile
        self.addConcealedTiles(TileList(tile))
        if deadEnd:
            self.lastSource = TileSource.DeadWall
        else:
            self.game.lastDiscard = None
            self.lastSource = TileSource.LivingWall
        return self.lastTile

    def removeConcealedTile(self, tile: Tile) -> None:
        """remove from my tiles"""
        assert not tile.isBonus, tile
        assert tile.__class__ == Tile
        if tile not in self._concealedTiles:
            raise ValueError(f'removeConcealedTile({tile!r}): tile not in concealed {self._concealedTiles!r}')
        self._concealedTiles.remove(tile)
        if tile is self.lastTile:
            self.lastTile = Tile.none
        self._hand = None

    def addConcealedTiles(self, tiles:'Tiles', animated:bool=False) -> None:  # pylint: disable=unused-argument
        """add to my tiles"""
        assert tiles
        for tile in tiles:
            if tile.isBonus:
                self._bonusTiles.append(tile)
            else:
                assert tile.isConcealed, f'{tile} data={tiles}'
                self._concealedTiles.append(tile)
        self._hand = None

    def syncHandBoard(self, adding:Optional[List['UITile']]=None) ->None:
        """virtual: synchronize display"""

    def colorizeName(self) ->None:
        """virtual: colorize Name on wall"""

    def getsFocus(self, unusedResults:Any=None) ->None:
        """virtual: player gets focus on his hand"""

    def __announcements(self) -> Set:
        """used to build the Hand"""
        return set('a') if self.originalCall else set()

    def makeTileKnown(self, tile:Tile) -> None:
        """used when somebody else discards a tile"""
        assert not self._concealedTiles[0].isKnown, f'{self._concealedTiles=}{self._concealedTiles[0].isKnown=}'
        self._concealedTiles[0] = tile
        self._hand = None

    def assertLastTile(self) ->None:
        """TODO: Remove again"""
        return
        # pylint: disable=unreachable
        if self.lastTile and self.lastTile.isKnown:
            if not self.lastTile.isBonus: # TODO: wirklich ausklammern?
                # pylint:disable=consider-using-f-string
                assert (self.lastTile in self._concealedTiles
                    or any(self.lastTile in x for x in self._exposedMelds)
                    or any(self.lastTile in x for x in self._concealedMelds)), \
                    'lastTile:{!r} conc:{!r} exp:{!r} concmelds:{!r}'.format(
                        self.lastTile, self._concealedTiles, self._exposedMelds, self._concealedMelds)

    def __computeHand(self) -> Hand:
        """return Hand for this player"""
        assert not (self._concealedMelds and self._concealedTiles)
        return Hand(
            self, melds=MeldList(self._exposedMelds + self._concealedMelds), unusedTiles=self._concealedTiles,
            bonusTiles=self._bonusTiles, lastTile=self.lastTile, lastMeld=self.lastMeld,
            lastSource=self.lastSource, announcements=self.__announcements())

    def _computeHandWithDiscard(self, discard:Tile) -> Hand:
        """what if"""
        lastSource = self.lastSource # TODO: recompute
        save = (self.lastTile, self.lastSource)
        try:
            self.lastSource = lastSource
            if discard:
                self.lastTile = discard
                self._concealedTiles.append(discard)
            return self.__computeHand()
        finally:
            self.lastTile, self.lastSource = save
            if discard:
                self._concealedTiles.pop(-1)

    def scoringString(self) -> str:
        """helper for HandBoard.__str__"""
        if self._concealedMelds:
            parts = [str(x) for x in self._concealedMelds + self._exposedMelds]
        else:
            parts = [str(self._concealedTiles)]
            parts.extend([str(x) for x in self._exposedMelds])
        parts.extend(str(x) for x in self._bonusTiles)
        return ' '.join(parts)

    def sortRulesByX(self, rules:List['UsedRule']) ->List['UsedRule']:
        """if this game has a GUI, sort rules by GUI order"""
        return rules

    def others(self) -> Generator['Player', None, None]:
        """a list of the other 3 players"""
        assert self.game
        return (x for x in self.game.players if x != self)

    def tileAvailable(self, tile:Tile, hand:Hand) -> int:
        """a count of how often tile might still appear in the game
        supposing we have hand"""
        lowerTile = tile.exposed
        upperTile = tile.concealed
        assert self.game
        visible = cast(IntDict, self.game.discardedTiles).count([lowerTile])
        if visible:
            if hand.lenOffset == 0 and self.game.lastDiscard and lowerTile is self.game.lastDiscard.exposed:
                # the last discarded one is available to us since we can claim
                # it
                visible -= 1
        visible += sum(cast(IntDict, x.visibleTiles).count([lowerTile, upperTile])
                       for x in self.others())
        visible += sum(x.exposed == lowerTile for x in hand.tiles)
        return 4 - visible

    def violatesOriginalCall(self, discard:Tile) ->bool:
        """called if discarding discard violates the Original Call"""
        if not self.originalCall or not self.mayWin:
            return False
        if self.lastTile and self.lastTile.exposed != discard.exposed:
            if Debug.originalCall and self.game:
                self.game.debug(
                    f'{self} would violate OC with {discard}, lastTile={self.lastTile}')
            return True
        return False


class PlayingPlayer(Player):  # pylint:disable=too-many-instance-attributes

    """a player in a computer game as opposed to a ScoringPlayer"""
    # too many public methods

    def __init__(self, game:Optional['PlayingGame'], name:str) ->None:
        self.sayable:Dict[Message, Union[bool, MeldList, Tuple[MeldList, Optional[Tile], Meld], None ]] = {}
        # recompute sayable for each move, use as cache
        self.voice:Optional['Voice']
        self.game:Optional['PlayingGame']
        super().__init__(game, name)

    def popupMsg(self, msg:'Message') ->None:
        """virtual: show popup on display"""

    def hidePopup(self) ->None:
        """virtual: hide popup on display"""

    def speak(self, txt:str) ->None:
        """only a visible playing player can speak"""

    def declaredMahJongg(self, concealed:MeldList, withDiscard:Optional[Tile], lastTile:Tile, lastMeld:Meld) ->None:
        """player declared mah jongg. Determine last meld, show concealed tiles grouped to melds"""
        assert self.game
        if Debug.mahJongg:
            self.game.debug(f'{self} declared MJ: concealed={concealed}, '
                            f'withDiscard={withDiscard}, lastTile={lastTile},lastMeld={lastMeld}')
            self.game.debug(f'  with hand being {self.hand}')
        melds = concealed[:]
        self.game.winner = self
        assert lastMeld in melds, (
            f'lastMeld {lastMeld} not in melds: concealed={self._concealedTiles}: '
            f'melds={melds} lastTile={lastTile} withDiscard={withDiscard}')
        if withDiscard:
            PlayingPlayer.addConcealedTiles(
                self,
                TileList(withDiscard))  # this should NOT invoke syncHandBoard
            if len(list(self.game.lastMoves(only=(Message.Discard, )))) == 1:
                self.lastSource = TileSource.East14th
            elif self.lastSource is not TileSource.RobbedKong:
                self.lastSource = TileSource.LivingWallDiscard
            # the last claimed meld is exposed
            melds.remove(lastMeld)
            lastTile = withDiscard.exposed
            lastMeld = lastMeld.exposed
            self._exposedMelds.append(lastMeld)
            for tileName in lastMeld:
                self.visibleTiles[tileName] += 1
        self.lastTile = lastTile or Tile.none
        self.lastMeld = lastMeld
        self._concealedMelds = melds
        self._concealedTiles = PieceList()
        self._hand = None
        if Debug.mahJongg:
            self.game.debug(f'  hand becomes {self.hand}')
            self._hand = None

    def __possibleChows(self) ->MeldList:
        """return a unique list of lists with possible claimable chow combinations"""
        if not self.game:
            return MeldList()
        if self.game.lastDiscard is None:
            return MeldList()
        exposedChows = [x for x in self._exposedMelds if x.isChow]
        if len(exposedChows) >= self.game.ruleset.maxChows:
            return MeldList()
        ladi = self.game.lastDiscard
        return (TileTuple(self.concealedTiles) + ladi).possibleChows(ladi)

    def __possibleKongs(self) ->MeldList:
        """return a unique list of lists with possible kong combinations"""
        if not self.game:
            return MeldList()
        kongs = MeldList()
        if self == self.game.activePlayer:
            # declaring a kong
            for tileName in sorted({x for x in self._concealedTiles if not x.isBonus}):
                if self._concealedTiles.count(tileName) == 4:
                    kongs.append(tileName.kong)
                elif self._concealedTiles.count(tileName) == 1 and \
                        tileName.exposed.pung in self._exposedMelds:
                    # the result will be an exposed Kong but the 4th tile
                    # came from the wall, so we use the form aaaA
                    kongs.append(tileName.kong.exposedClaimed)
        if self.game.lastDiscard:
            # claiming a kong
            discardedTile = self.game.lastDiscard.concealed
            if self._concealedTiles.count(discardedTile) == 3:
                # discard.kong.concealed is aAAa but we need AAAA
                kongs.append(Meld(discardedTile * 4))
        return kongs

    def __maySayChow(self, unusedMove:'Move') ->MeldList:
        """return answer arguments for the server if calling chow is possible.
        returns the meld to be completed"""
        if not self.game:
            return MeldList()
        return self.__possibleChows() if self == self.game.nextPlayer() else MeldList()

    def __maySayPung(self, unusedMove:'Move') -> MeldList:
        """return answer arguments for the server if calling pung is possible.
        returns the meld to be completed"""
        if not self.game:
            return MeldList()
        lastDiscard = self.game.lastDiscard
        if lastDiscard:
            assert lastDiscard.isConcealed, lastDiscard
            if self.concealedTiles.count(lastDiscard) >= 2:
                return MeldList([lastDiscard.pung])
        return MeldList()

    def __maySayKong(self, unusedMove:'Move') ->MeldList:
        """return answer arguments for the server if calling or declaring kong is possible.
        returns the meld to be completed or to be declared"""
        return self.__possibleKongs()

    def __maySayMahjongg(self, move:'Move') ->Optional[Tuple[MeldList, Optional[Tile], Optional[Meld]]]:
        """return answer arguments for the server if calling or declaring Mah Jongg is possible"""
        game = self.game
        assert game
        if move.message == Message.DeclaredKong:  # type: ignore
            assert move.meld
            withDiscard = move.meld[0].concealed
        elif move.message == Message.AskForClaims:  # type: ignore
            withDiscard = game.lastDiscard
        else:
            withDiscard = None
        hand = self._computeHandWithDiscard(withDiscard)
        if hand.won:
            if Debug.robbingKong:
                if move.message == Message.DeclaredKong:  # type: ignore
                    game.debug(f'{self} may rob the kong from {move.player}/{move.exposedMeld}')
            if Debug.mahJongg:
                game.debug(f'{self} may say MJ:{list(x for x in game.players)}, active={game.activePlayer}')
                game.debug(f'  with hand {hand}')
            return MeldList(x for x in hand.melds if not x.isDeclared), withDiscard, hand.lastMeld
        return None

    def __maySayOriginalCall(self, unusedMove:'Move') ->bool:
        """return True if Original Call is possible"""
        assert self.game
        for tileName in sorted(set(self.concealedTiles)):
            newHand = self.hand - tileName
            if newHand.callingHands:
                if Debug.originalCall:
                    self.game.debug(
                        f'{self} may say Original Call by discarding {tileName} from {self.hand}')
                return True
        return False

    __sayables:Dict[Any, Callable] = {
        Message.Pung: __maySayPung,
        Message.Kong: __maySayKong,
        Message.Chow: __maySayChow,
        Message.MahJongg: __maySayMahjongg,
        Message.OriginalCall: __maySayOriginalCall}

    def computeSayable(self, move:'Move', answers:List['ClientMessage']) ->None:
        """find out what the player can legally say with this hand"""
        self.sayable = {}
        for message in Message.defined.values():
            if message in answers and message in self.__sayables:
                self.sayable[message] = self.__sayables[message](self, move)
            else:
                self.sayable[message] = True

    def maybeDangerous(self, msg:Message) ->MeldList:
        """could answering with msg lead to dangerous game?
        If so return a list of resulting melds
        where a meld is represented by a list of 2char strings"""
        result = MeldList()
        if msg in (Message.Chow, Message.Pung, Message.Kong):
            for meld in cast(MeldList, self.sayable[msg]):
                if self.mustPlayDangerous(meld):
                    result.append(meld)
        return result

    def hasConcealedTiles(self, tiles:'Tiles', within:Optional['Tiles'] =None) -> bool:
        """do I have those concealed tiles?"""
        if within is None:
            within = self._concealedTiles
        within = PieceList(within[:])  # type: ignore
        for tile in tiles:
            if tile not in within:
                return False
            within.remove(tile)
        return True

    def showConcealedTiles(self, tiles:TileTuple, show:bool=True) ->None:
        """show or hide tiles"""
        assert isinstance(tiles, TileTuple), repr(tiles)
        assert self.game
        if not self.game.playOpen and self != self.game.myself:
            assert len(tiles) <= len(self._concealedTiles), \
                f'{self}: showConcealedTiles {tiles}, we have only {self._concealedTiles}'
            for tileName in tiles:
                src, dst = (Tile.unknown, tileName) if show else (
                    tileName, Tile.unknown)
                assert src != dst, f'{self=} {src=} {dst=} {tiles=} {self._concealedTiles=}'
                if src not in self._concealedTiles:
                    logException(f'{self}: showConcealedTiles({tiles}): {src} not in {self._concealedTiles}.')
                idx = self._concealedTiles.index(src)
                self._concealedTiles[idx] = dst
            self._hand = None
            self.syncHandBoard()

    def showConcealedMelds(self, concealedMelds:MeldList,
        ignoreDiscard:Optional[Tile] =None) ->Optional[Tuple[str, ...]]:
        """the server tells how the winner shows and melds his
        concealed tiles. In case of error, return message and arguments"""
        for meld in concealedMelds:
            for tile in meld:
                if tile == ignoreDiscard:
                    ignoreDiscard = None
                else:
                    if tile not in self._concealedTiles:
                        msg = i18nE(
                            '%1 claiming MahJongg: She does not really have tile %2')
                        return msg, self.name, tile
                    self._concealedTiles.remove(tile)
            if meld.isConcealed and not meld.isKong:
                self._concealedMelds.append(meld)
            else:
                self._exposedMelds.append(meld)
        if self._concealedTiles:
            msg = i18nE(
                '%1 claiming MahJongg: She did not pass all concealed tiles to the server')
            return msg, self.name
        self._hand = None
        return None

    def getsRobbed(self, tile:Tile) -> None:
        """used for robbing the kong from this player"""
        if Debug.robbingKong:
            logDebug(f'robbed {tile} from {self}')
        assert tile.isConcealed
        tile = tile.exposed
        for meld in self._exposedMelds:
            if tile in meld:
                # FIXME: document if and where Player is updated meld = meld.without(tile)
                self.visibleTiles[tile] -= 1
                break
        else:
            # TODO: should we somehow show an error and continue?
            raise ValueError(f'getsRobbed: no meld found with {tile}')
        assert self.game
        self.game.lastDiscard = tile.concealed
        self.lastTile = Tile.none  # our lastTile has just been robbed
        self._hand = None

    def robsTile(self) -> None:
        """True if the player is robbing a tile"""
        if Debug.robbingKong:
            logDebug(f'{self} robs a tile')
        self.lastSource = TileSource.RobbedKong

    def scoreMatchesServer(self, score:Optional[str]) -> bool:
        """do we compute the same score as the server does?"""
        if score is None:
            return True
        if any(not x.isKnown for x in self._concealedTiles):
            return True
        if str(self.hand) == score:
            return True
        if self.game:
            self.game.debug(f'{self} localScore:{self.hand}')
            self.game.debug(f'{self} serverScore:{score}')
            logWarning(
                f'Game {self.game.seed}: client and server disagree about scoring, see logfile for details')
        return False

    def mustPlayDangerous(self, exposing:Optional[Meld] =None) -> bool:
        """]
        True if the player has no choice, otherwise False.

        @param exposing: May be a meld which will be exposed before we might
        play dangerous.
        @type exposing: L{Meld}
        @rtype: C{Boolean}
        """
        assert self.game
        if self == self.game.activePlayer and exposing and len(exposing) == 4:
            # declaring a kong is never dangerous because we get
            # an unknown replacement
            return False
        afterExposed = [x.exposed for x in self._concealedTiles]
        if exposing:
            exposingTiles = TileList(exposing)
            if self.game.lastDiscard:
                # if this is about claiming a discarded tile, ignore it
                # the player who discarded it is responsible
                exposingTiles.remove(self.game.lastDiscard)
            for tile in exposingTiles:
                if tile.exposed in afterExposed:
                    # the "if" is needed for claimed pung
                    afterExposed.remove(tile.exposed)
        return all(self.game.dangerousFor(self, x) for x in afterExposed)

    def exposeMeld(self, meldTiles: 'Tiles', calledTile:Optional['Piece']=None) -> Meld:
        """exposes a meld with meldTiles: removes them from concealedTiles,
        adds the meld to exposedMelds and returns it
        calledTile: we got the last tile for the meld from discarded, otherwise
        from the wall"""
        self.assertLastTile()
        game = self.game
        assert game
        game.activePlayer = self
        allMeldTiles = list(meldTiles)
        if calledTile:
            assert isinstance(calledTile, Tile), calledTile
            allMeldTiles.append(calledTile)
        if len(allMeldTiles) == 4 and allMeldTiles[0].isExposed:
            tile0 = allMeldTiles[0].exposed
            # we are adding a 4th tile to an exposed pung
            self._exposedMelds = MeldList(
                x for x in self._exposedMelds if x != tile0.pung)
            meld = tile0.kong
            if allMeldTiles[3] not in self._concealedTiles:
                game.debug(
                    f't3 {allMeldTiles[3]} not in conc {self._concealedTiles}')
            self._concealedTiles.remove(allMeldTiles[3])
            self.visibleTiles[tile0] += 1
        else:
            allMeldTiles = sorted(allMeldTiles)  # needed for Chow
            meld = Meld(allMeldTiles)
            for meldTile in meldTiles:
                self._concealedTiles.remove(meldTile)
            for meldTile in allMeldTiles:
                self.visibleTiles[meldTile.exposed] += 1
            meld = meld.exposedClaimed if calledTile else meld.declared
        assert meld
        if self.lastTile in allMeldTiles:
            self.lastTile = self.lastTile.exposed
        self._exposedMelds.append(meld)
        self._hand = None
        game.computeDangerous(self)
        return meld

    def findDangerousTiles(self) -> None:
        """update the list of dangerous tile"""
        assert self.game
        pName = self.localName
        dangerous:List[Tuple[Set[Tile], str]] = []
        expMeldCount = len(self._exposedMelds)
        if expMeldCount >= 3:
            if all(x in elements.greenHandTiles for x in self.visibleTiles):
                dangerous.append((elements.greenHandTiles,
                                  i18n('Player %1 has 3 or 4 exposed melds, all are green', pName)))
            group = list(self.visibleTiles.keys())[0].group
            assert group.islower(), self.visibleTiles
            if group in Tile.colors:
                if all(x.group == group for x in self.visibleTiles):
                    suitTiles = {Tile(group, x) for x in Tile.numbers}
                    if cast(IntDict, self.visibleTiles).count(suitTiles) >= 9:
                        dangerous.append(
                            (suitTiles, i18n('Player %1 may try a True Color Game', pName)))
                elif all(x.value in Tile.terminals for x in self.visibleTiles):
                    dangerous.append((elements.terminals,
                                      i18n('Player %1 may try an All Terminals Game', pName)))
        if expMeldCount >= 2:
            windMelds = sum(self.visibleTiles[x] >= 3 for x in elements.winds)
            dragonMelds = sum(
                self.visibleTiles[x] >= 3 for x in elements.dragons)
            windsDangerous = dragonsDangerous = False
            if windMelds + dragonMelds == expMeldCount and expMeldCount >= 3:
                windsDangerous = dragonsDangerous = True
            windsDangerous = windsDangerous or windMelds >= 3
            dragonsDangerous = dragonsDangerous or dragonMelds >= 2
            if windsDangerous:
                dangerous.append(
                    ({x for x in elements.winds if x not in self.visibleTiles},
                     i18n('Player %1 exposed many winds', pName)))
            if dragonsDangerous:
                dangerous.append(
                    ({x for x in elements.dragons if x not in self.visibleTiles},
                     i18n('Player %1 exposed many dragons', pName)))
        self.dangerousTiles = dangerous
        if dangerous and Debug.dangerousGame:
            self.game.debug(f'dangerous:{dangerous}')



# -*- coding: utf-8 -*-
"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

Here we define classes useful for tree views
"""

from typing import Any, List, Sequence, Optional, overload, Union, TYPE_CHECKING, cast

from qt import QAbstractItemModel, QModelIndex, QPersistentModelIndex

if TYPE_CHECKING:
    from qt import QObject


class TreeItem:

    """generic class for items in a tree"""

    def __init__(self, content: Any) ->None:
        self.__rawContent = content
        self.parent:Optional['TreeItem'] = None
        self.children:List['TreeItem'] = []

    def insert(self, row:int, child:'TreeItem') ->'TreeItem':
        """add a new child to this tree node"""
        assert isinstance(child, TreeItem)
        child.parent = self
        self.children.insert(row, child)
        return child

    def remove(self) ->None:
        """remove this item from the model and the database.
        This is an abstract method."""
        raise TypeError(
            'cannot remove this TreeItem. We should never get here.')

    def child(self, row:int) ->'TreeItem':
        """return a specific child item"""
        return self.children[row]

    def childCount(self) ->int:
        """how many children does this item have?"""
        return len(self.children)

    def content(self, column:int) ->Any:
        """content held by this item"""
        raise NotImplementedError("Virtual Method")

    def row(self) ->int:
        """the row of this item in parent"""
        if self.parent:
            return self.parent.children.index(self)
        return 0

    def columnCount(self) ->int:
        """Always return 1. Is 1 always correct? No, inherit from RootItem FIXME"""
        return 1

    @property
    def raw(self) ->Any:
        """make it read only"""
        return self.__rawContent


class RootItem(TreeItem):

    """an item for header data"""

    def content(self, column:int) ->Any:
        """content held by this item"""
        return self.raw[column]


class TreeModel(QAbstractItemModel):

    """a basic class for Kajongg tree views"""

    def __init__(self, parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        self.rootItem:Optional[RootItem] = None

    def columnCount(self, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """how many columns does this node have?"""
        return self.itemForIndex(cast(QModelIndex, parent)).columnCount()

    def rowCount(self, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """how many items?"""
        if parent.isValid() and parent.column():
            # all children have col=0 for parent
            return 0
        return self.itemForIndex(cast(QModelIndex, parent)).childCount()

    def index(self, row:int, column:int, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->QModelIndex:
        """generate an index for this item"""
        if self.rootItem is None:
            return QModelIndex()
        if (row < 0
                or column < 0
                or row >= self.rowCount(parent)
                or column >= self.columnCount(parent)):
            return QModelIndex()
        parentItem = self.itemForIndex(cast(QModelIndex, parent))
        assert parentItem
        item = parentItem.child(row)
        if item:
            return self.createIndex(row, column, item)
        return QModelIndex()

    @overload  # type:ignore[override]
    def parent(self) ->'QObject': ...
    @overload
    def parent(self, index:Union[QModelIndex,QPersistentModelIndex]) ->QModelIndex: ...

    def parent(self, index:Optional[Union[QModelIndex,QPersistentModelIndex]]=None) ->Union[QModelIndex, 'QObject']:
        """find the parent for index"""
        assert isinstance(index, QModelIndex)  # we never use parent() for getting QObject
        if not index.isValid():
            return QModelIndex()
        childItem = self.itemForIndex(index)
        if childItem:
            parentItem = childItem.parent
            if parentItem:
                if parentItem != self.rootItem:
                    grandParentItem = parentItem.parent
                    if grandParentItem:
                        row = grandParentItem.children.index(parentItem)
                        return self.createIndex(row, 0, parentItem)
        return QModelIndex()

    def itemForIndex(self, index:QModelIndex) ->TreeItem:
        """return the item at index"""
        if index.isValid():
            item = cast(TreeItem, index.internalPointer())
            if item:
                return item
        assert self.rootItem
        return self.rootItem

    def insertRows(self, position:int, items:Sequence[TreeItem],  # type:ignore[override]
        parent:QModelIndex=QModelIndex()) ->bool:
        """inserts items into the model"""
        parentItem = self.itemForIndex(parent)
        self.beginInsertRows(parent, position, position + len(items) - 1)
        for row, item in enumerate(items):
            parentItem.insert(position + row, item)
        self.endInsertRows()
        return True

    def removeRows(self, position:int, rows:int=1,
        parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->bool:
        """reimplement QAbstractItemModel.removeRows"""
        assert isinstance(parent, QModelIndex)
        self.beginRemoveRows(parent, position, position + rows - 1)
        parentItem = self.itemForIndex(parent)
        for row in parentItem.children[position:position + rows]:
            row.remove()
        parentItem.children = parentItem.children[
            :position] + parentItem.children[
                position + rows:]
        self.endRemoveRows()
        return True



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import TYPE_CHECKING, Optional

from common import Internal

from qt import Qt, QWidget, QSlider, QHBoxLayout, QLabel
from qt import QVBoxLayout, QSpacerItem, QSizePolicy, QCheckBox

from kde import KConfigDialog
from mi18n import i18n, i18nc

from statesaver import StateSaver
from tilesetselector import TilesetSelector
from backgroundselector import BackgroundSelector

if TYPE_CHECKING:
    from qt import QKeyEvent

__all__ = ['ConfigDialog']


class PlayConfigTab(QWidget):

    """Display Config tab"""

    def __init__(self, parent:QWidget) ->None:
        super().__init__(parent)
        self.setupUi()

    def setupUi(self) ->None:
        """layout the window"""
        self.setContentsMargins(0, 0, 0, 0)
        vlayout = QVBoxLayout(self)
        vlayout.setContentsMargins(0, 0, 0, 0)
        sliderLayout = QHBoxLayout()
        self.kcfg_showShadows = QCheckBox(i18n('Show tile shadows'), self)
        self.kcfg_showShadows.setObjectName('kcfg_showShadows')
        self.kcfg_rearrangeMelds = QCheckBox(
            i18n('Rearrange undisclosed tiles to melds'), self)
        self.kcfg_rearrangeMelds.setObjectName('kcfg_rearrangeMelds')
        self.kcfg_showOnlyPossibleActions = QCheckBox(i18n(
            'Show only possible actions'))
        self.kcfg_showOnlyPossibleActions.setObjectName(
            'kcfg_showOnlyPossibleActions')
        self.kcfg_propose = QCheckBox(i18n('Propose what to do'))
        self.kcfg_propose.setObjectName('kcfg_propose')
        self.kcfg_animationSpeed = QSlider(self)
        self.kcfg_animationSpeed.setObjectName('kcfg_animationSpeed')
        self.kcfg_animationSpeed.setOrientation(Qt.Orientation.Horizontal)
        self.kcfg_animationSpeed.setSingleStep(1)
        lblSpeed = QLabel(i18n('Animation speed:'))
        lblSpeed.setBuddy(self.kcfg_animationSpeed)
        sliderLayout.addWidget(lblSpeed)
        sliderLayout.addWidget(self.kcfg_animationSpeed)
        self.kcfg_useSounds = QCheckBox(i18n('Use sounds if available'), self)
        self.kcfg_useSounds.setObjectName('kcfg_useSounds')
        self.kcfg_uploadVoice = QCheckBox(i18n(
            'Let others hear my voice'), self)
        self.kcfg_uploadVoice.setObjectName('kcfg_uploadVoice')
        pol = QSizePolicy()
        pol.setHorizontalPolicy(QSizePolicy.Policy.Expanding)
        pol.setVerticalPolicy(QSizePolicy.Policy.Expanding)
        spacerItem = QSpacerItem(
            20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        vlayout.addWidget(self.kcfg_showShadows)
        vlayout.addWidget(self.kcfg_rearrangeMelds)
        vlayout.addWidget(self.kcfg_showOnlyPossibleActions)
        vlayout.addWidget(self.kcfg_propose)
        vlayout.addWidget(self.kcfg_useSounds)
        vlayout.addWidget(self.kcfg_uploadVoice)
        vlayout.addLayout(sliderLayout)
        vlayout.addItem(spacerItem)
        self.setSizePolicy(pol)
        self.retranslateUi()

    def retranslateUi(self) ->None:
        """translate to current language"""


class ConfigDialog(KConfigDialog):

    """configuration dialog with several pages"""

    def __init__(self, parent:QWidget, name:str) ->None:
        assert Internal.Preferences
        KConfigDialog.__init__(
            self, parent, name, Internal.Preferences)
        StateSaver(self)
        self.pages = [
            self.addPage(
                PlayConfigTab(self),
                i18nc('kajongg', 'Play'), "arrow-right"),
            self.addPage(
                TilesetSelector(self),
                i18n("Tiles"), "games-config-tiles"),
            self.addPage(
                BackgroundSelector(self),
                i18n("Backgrounds"), "games-config-background")]

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """The four tabs can be selected with CTRL-1 .. CTRL-4"""
        if not event:
            return
        mod = event.modifiers()
        key = chr(event.key() % 128)
        if Qt.KeyboardModifier.ControlModifier | mod and key in '123456789'[:len(self.pages)]:
            self.setCurrentPage(self.pages[int(key) - 1])
            return
        super().keyPressEvent(event)



# -*- coding: utf-8 -*-

"""
 Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from itertools import chain
from types import GeneratorType
from typing import Dict, Any, Optional, cast, Union, Tuple, Sequence, Type
from typing import Iterator, List, TYPE_CHECKING, Generator, Iterable
from log import logException
from mi18n import i18n, i18nc
from common import IntDict, ReprMixin, id4
from wind import Wind, East, South, West, North

if TYPE_CHECKING:
    from rule import Rule, Ruleset
    from hand import Hand
    from uitile import UITile

class Tile(ReprMixin):  # pylint:disable=too-many-instance-attributes

    """
    A single tile, represented as a string of length 2.

    Always True:
      - only for suits: tile.group + chr(tile.value + 48) == str(tile)
      - tile.group + tile.char == str(tile)
      - Tile(tile) is tile
      - Tile(tile.group, tile.value) is tile

    Tile() accepts
      - another Tile
      - a string, length 2
      - two args: a char and either a char or an int -1..11

    group is a char: b=bonus w=wind d=dragon X=unknown
    value is
        1..9 for real suit tiles
        -1/0/10/11 for usage in AI
        Wind for winds and boni
        bgr for dragons
    """
    cache:Dict[Any, 'Tile'] = {}
    hashTable = 'XxxxXyxyDbdbDgdgDrdrWeweWswsWw//wwWnwn' \
                'S/s/S0s0S1s1S2s2S3s3S4s4S5s5S6s6S7s7S8s8S9s9S:s:S;s;' \
                'B/b/B0b0B1b1B2b2B3b3B4b4B5b5B6b6B7b7B8b8B9b9B:b:B;b;' \
                'C/c/C0c0C1c1C2c2C3c3C4c4C5c5C6c6C7c7C8c8C9c9C:c:C;c;' \
                'fefsfwfnyeysywyn'
    # the // is needed as separator between too many w's
    # intelligence.py will define Tile('b0') or Tile('s:')

    unknown:'Tile'
    unknownStr:str = 'Xy'
    none:'Tile'
    noneStr:str = 'Xx'

    # Groups:
    hidden = 'x'
    stone = 's'
    bamboo = 'b'
    character = 'c'
    colors = stone + bamboo + character
    wind = 'w'
    dragon = 'd'
    honors = wind + dragon
    flower = 'f'
    season = 'y'
    boni = flower + season

    # Values:
    dragons = 'bgr'
    white, green, red = tuple(dragons)
    winds = 'eswn'
    numbers = range(1, 10)
    terminals = list([1, 9])
    minors = range(2, 9)
    majors = list(dragons) + list(winds) + terminals

    def __new__(cls, *args: Any) ->'Tile':
        if len(args) == 1:
            if  type(args[0]) is Tile:  # pylint: disable=unidiomatic-typecheck
                return args[0]
            if  type(args[0]) is Piece:  # pylint: disable=unidiomatic-typecheck
                args = tuple([args[0].name2()])
# FIXME: wirklich list als key?
# lieber arg1, arg2 und fuer dict und _build daraus eines machen
        try:
            return cls.cache[args]
        except KeyError:
            return cls._build(*args)

    @classmethod
    def _build(cls, *args: Any) ->'Tile':
        """build a new Tile object out of args"""

        result = super().__new__(cls)
        result.setUp(args)
        return result

    def setUp(self, args:Any) ->None:  # pylint:disable=too-many-statements, too-many-branches
        """Initialize"""

        self.group:str
        self.char:str
        self.value:Union[str, int, Wind]

        # parse args
        arg1:Union[str, int, Wind]
        if isinstance(args, Tile):
            group, arg1 = args.group, args.char
        elif len(args) == 1:
            if isinstance(args[0], Tile):
                group = args[0].group
                arg1 = args[0].value
            else:
                assert isinstance(args[0], str)
                try:
                    group, arg1 = args[0]  # type:ignore[misc]
                    # mypy: unpacking a string is disallowed
                except ValueError:
                    logException(f'cannot make a tile from {args}')
        else:
            group, arg1 = args
        self.group = group

        # set attributes depending only on group
        self.isKnown = self.group != 'X'
        self.lowerGroup = self.group.lower()
        self.isExposed = self.group == self.lowerGroup
        self.isConcealed = not self.isExposed
        self.isBonus = self.group in Tile.boni
        self.isDragon = self.lowerGroup == Tile.dragon
        self.isWind = self.lowerGroup == Tile.wind
        self.isHonor = self.isDragon or self.isWind

        self.char, self.value = self.parse_arg1(arg1)

        self.key:int
        try:
            self.key = 1 + self.hashTable.index(Tile.__str__(self)) // 2
        except ValueError:
            logException(f'{self} is not a valid tile string')

        self.isNumber = False
        self.isTerminal = False
        self.isReal = False

        if self.isBonus or self.isDragon or self.isWind:
            self.isReal = True
        elif self.isKnown:
            self.isNumber = True
            self.isTerminal = self.value in Tile.terminals
            self.isReal = self.value in Tile.numbers
        self.isMajor = self.isHonor or self.isTerminal
        self.isMinor = self.isKnown and not self.isMajor

        if self.__class__ is Tile:
            Tile._storeInCache(self)

        self.exposed:'Tile'
        self.concealed:'Tile'
        self.swapped:'Tile'
        self.knitted3:'Meld'
        self.claimedKong:'Meld'
        self.single:'Meld'
        self.pair:'Meld'
        self.pung:'Meld'
        self.chow:'Meld'
        self.kong:'Meld'
        self.nextForChow:'Tile'
        if self.__class__ is Tile:
            # a Piece may change
            self._fixed = True

        object.__setattr__(
            self,
            'exposed',
            Tile(self.name2()) if not self.isKnown else Tile(self.group.lower(), self.char))
        object.__setattr__(self, 'concealed',
                        Tile(self.name2()) if not self.isKnown or self.isBonus
                        else Tile(self.group.upper(), self.char))
        object.__setattr__(
            self,
            'swapped',
            self.exposed if self.isConcealed else self.concealed)
        if isinstance(self.value, int):
            if 0 <= self.value <= 11:
                object.__setattr__(
                    self,
                    'prevForChow',
                    Tile(self.group,
                         self.value - 1))
            if -1 <= self.value <= 10:
                object.__setattr__(
                    self,
                    'nextForChow',
                    Tile(
                        self.group,
                        self.value +
                        1))

    def parse_arg1(self, arg1:Any) ->Tuple[str, Union[str, int, Wind]]:
        """set the interpreted Tile.value (str, Wind, int)"""
        if isinstance(arg1, Wind):
            char = arg1.char.lower()
        elif isinstance(arg1, int):
            char = chr(arg1 + 48)
        else:
            char = arg1
        value:Union[str, int, Wind] = char  # default
        if self.isWind or self.isBonus:
            if isinstance(arg1, Wind):
                value = arg1
            else:
                value = Wind(arg1)
        elif self.isDragon:
            assert char in 'gbr', arg1
        elif self.group != 'X':
            if isinstance(char, int):
                value = char
            else:
                value = ord(char) - 48
        return char, value

    @classmethod
    def _storeInCache(cls, result:'Tile') ->None:
        """Put the new tile into the cache"""
        for key in (
                result, (str(result),), (result.group, result.value),
                (result.group, result.char)):
            cls.cache[key] = result

        existing = list([x for x in cls.cache.values() if x.key == result.key]) # pylint: disable=consider-using-generator
        existingIds = {id4(x) for x in existing}
        assert len(existingIds) == 1, (f"cls is {cls.__name__} new is:{repr(result)} "
                                       f"existing are: {','.join(repr(x) for x in existing)} "
                                       f"keys are:{','.join(repr(x) for x in cls.cache if x == 1)}")

    def name2(self) ->str:
        """__str__ might be changed by a subclass"""
        return self.group + self.char

    def __repr__(self) ->str:
        """ReprMixin does not seem to work on str subclass"""
        return ReprMixin.__repr__(self)

    def __str__(self) ->str:
        return self.group + self.char

    def __hash__(self) ->int:
        return self.key

    def __mul__(self, other:int) ->'TileList':
        return TileList([self] * other)

    def __imul__(self, other:int) ->'TileList':
        return TileList([self] * other)

    def __bool__(self) ->bool:
        return self.isKnown

    def meld(self, size:int) ->'Meld':
        """return a meld of size. Those attributes are set
        in Meld.cacheMeldsInTiles"""
        _ = self
        if isinstance(_, Piece):
            _ = Tile(_)
        return getattr(_, ('single', 'pair', 'pung', 'kong')[size - 1])

    def groupName(self) ->str:
        """the name of the group this tile is of"""
        names = {
            Tile.hidden: i18nc('kajongg', 'hidden'),
            Tile.stone: i18nc('kajongg', 'stone'),
            Tile.bamboo: i18nc('kajongg', 'bamboo'),
            Tile.character: i18nc('kajongg', 'character'),
            Tile.wind: i18nc('kajongg', 'wind'),
            Tile.dragon: i18nc('kajongg', 'dragon'),
            Tile.flower: i18nc('kajongg', 'flower'),
            Tile.season: i18nc('kajongg', 'season')}
        return names[self.lowerGroup]

    def valueName(self) ->str:
        """the name of the value this tile has"""
        names = {
            'y': i18nc('kajongg', 'tile'),
            Tile.white: i18nc('kajongg', 'white'),
            Tile.red: i18nc('kajongg', 'red'),
            Tile.green: i18nc('kajongg', 'green'),
            East: i18nc('kajongg', 'East'),
            South: i18nc('kajongg', 'South'),
            West: i18nc('kajongg', 'West'),
            North: i18nc('kajongg', 'North')}
        for idx in Tile.numbers:
            names[idx] = chr(idx + 48)
        return names[self.value]

    def name(self) ->str:
        """return name of a single tile"""
        if isinstance(self.value, Wind):
            result = {
                East: i18n('East Wind'),
                South: i18n('South Wind'),
                West: i18n('West Wind'),
                North: i18n('North Wind')}[self.value]
        else:
            result = i18nc('kajongg tile name', '{group} {value}')
        return result.format(value=self.valueName(), group=self.groupName())

    def __lt__(self, other:Any) ->bool:
        """needed for sort"""
        return self.key < other.key

    def change_name(self, element:'Tile') ->'Tile':
        """FIXME: should go away when Tile/Piece is done"""
        assert element.__class__ is Tile, repr(element)
        if element is Tile.none:
            return self
        assert element.__class__ is Tile, repr(element)
        return cast('Tile', element)

    def __eq__(self, other:Any) ->bool:
        if isinstance(other, Tile):
            return self.name2() == other.name2()
        return object.__eq__(self, other)

    def cacheMelds(self) ->None:
        """fill meld cache"""
        occ = elements.occurrence[self]
        self.single = Meld(self)
        self.concealed.single = Meld(self.concealed)
        if occ > 1:
            self.pair = Meld(self * 2)
            self.concealed.pair = Meld(self.concealed * 2)
            if occ > 2:
                self.pung = Meld(self * 3)
                self.concealed.pung = Meld(self.concealed * 3)
                if self.value in range(1, 8):
                    self.chow = Meld(
                        [self,
                         self.nextForChow,
                         self.nextForChow.nextForChow])
                    self.concealed.chow = Meld(
                        [self.concealed,
                         self.concealed.nextForChow,
                         self.concealed.nextForChow.nextForChow])
                if self.value in range(1, 10):
                    self.knitted3 = Meld(
                        [Tile(x, self.value) for x in Tile.colors])
                    self.concealed.knitted3 = Meld(
                        [Tile(x, self.value).concealed for x in Tile.colors])
                if occ > 3:
                    self.kong = Meld(self * 4)
                    self.claimedKong = Meld(
                        [self, self, self, self.concealed])
                    self.concealed.kong = Meld(self.concealed * 4)

class Tiles:

    """a Mixin for TileList and TileTuple"""


    tileClass:Type

    def __init__(self, newContent:Any=None) ->None:  # pylint:disable=unused-argument
        ...

    @classmethod
    def _parseArgs(cls, iterable:Any) ->Sequence:
        """flatten any args into a nice sequence"""
        if isinstance(iterable, str):
            memberList = [cls.tileClass(iterable[x:x + 2])
                       for x in range(0, len(iterable), 2)]
        elif isinstance(iterable, Tile) or iterable.__class__.__name__ == 'UITile':
            memberList = [iterable]
        elif iterable is None:
            memberList = []
        else:
            memberList = []
            for member in iterable:
                if isinstance(member, cls.tileClass):
                    memberList.append(member)
                elif isinstance(member, str):
                    memberList.extend(cls.tileClass(member[x:x + 2])
                               for x in range(0, len(member), 2))
                elif hasattr(member, '__iter__'):
                    memberList.extend(member)
                elif isinstance(member, Tile): # FIXME: remove again
                    memberList.append(member)
                else:
                    raise ValueError(
                        f'{cls.__name__}() accepts only {cls.tileClass.__name__} and str but got {repr(member)}')
        return memberList

    def sorted(self) ->'Tiles':
        """sort(TileList) would not keep TileList type"""
        return self.__class__(sorted(self))

    def possibleChows(self, tile:Tile) ->'MeldList':
        """return my chows with tileName"""
        if tile not in self:
            return []  # type:ignore[return-value]
        if tile.lowerGroup not in Tile.colors:
            return []  # type:ignore[return-value]
        assert isinstance(tile.value, int)
        group = tile.group
        values = {x.value for x in self if x.group == group}
        chows:'MeldList' = []  # type:ignore[assignment]
        for offsets in [(0, 1, 2), (-2, -1, 0), (-1, 0, 1)]:
            subset = {tile.value + x for x in offsets}
            if subset <= values:
                chow = self.__class__(self.tileClass(group, x) for x in sorted(subset))
                if chow not in chows:
                    chows.append(chow)  # type:ignore[arg-type]
        return chows

    def __str__(self) ->str:
        """the content"""
        return str(''.join(str(x) for x in self))

    def __repr__(self) ->str:
        """for debugging"""
        return f"{self.__class__.__name__}_{id4(self)}({','.join(repr(x) for x in self)})"

    def __len__(self) ->int:
        """just to make this clear to mypy"""
        return 0

    def __iter__(self) -> Iterator[Tile]:
        """just to make this clear to mypy"""
        return iter(self)


class TileList(list, Tiles):

    """a list that can only hold tiles"""

    tileClass = Tile

    def __init__(self, newContent:Any=None) ->None:
        list.__init__(self)
        Tiles.__init__(self, newContent)
        self.extend(self._parseArgs(newContent))
        if self:
            self.isRest = True

    def __add__(self, other:Any) ->'TileList':
        result = TileList(self)
        result.extend(self._parseArgs(other))
        return result


class TileTuple(tuple, Tiles):

    """a tuple that can only hold tiles"""

    tileClass = Tile

    def __new__(cls, iterable:Any=None) ->'TileTuple':
        result = tuple.__new__(cls, cls._parseArgs(iterable))
        result.isRest = True
        result._hash = result._compute_hash()
        return result

    def __init__(self, newContent:Any=None) ->None:  # pylint: disable=unused-argument
        tuple.__init__(self)
        Tiles.__init__(self)
        self._hash:int
        self.isRest:bool

    def _compute_hash(self) ->int:
        """usable for sorting"""
        result = 0
        factor = len(Tile.hashTable) // 2
        for tile in self:
            result = result * factor + tile.key
        return result

    def __hash__(self) ->int:
        return self._hash

    def __add__(self, other:Any) ->'TileTuple':  # type:ignore[override]
        return TileTuple(TileList(self) + self._parseArgs(other))


class Elements:

    """represents all elements"""
    # pylint: disable=too-many-instance-attributes
    # too many attributes

    def __init__(self) ->None:
        self.occurrence:Dict[Tile, int] = IntDict()  # key: db, s3 etc. value: occurrence
        self.winds = {Tile(Tile.wind, x) for x in Tile.winds}
        self.wINDS = {x.concealed for x in self.winds}
        self.dragons = {Tile(Tile.dragon, x) for x in Tile.dragons}
        self.dRAGONS = {x.concealed for x in self.dragons}
        self.honors = self.winds | self.dragons
        self.hONORS = self.wINDS | self.dRAGONS
        self.terminals = {Tile(x, y)
                          for x in Tile.colors for y in Tile.terminals}
        self.tERMINALS = {x.concealed for x in self.terminals}
        self.majors = self.honors | self.terminals
        self.mAJORS = self.hONORS | self.tERMINALS
        self.greenHandTiles = {
            Tile(Tile.bamboo, x)
            for x in '23468'} | {Tile(Tile.dragon, Tile.green)}
        self.minors = {Tile(x, y) for x in Tile.colors for y in Tile.minors}
        for tile in self.majors:
            self.occurrence[tile] = 4
        for tile in self.minors:
            self.occurrence[tile] = 4
        for bonus in Tile.boni:
            for _ in Tile.winds:
                self.occurrence[Tile(bonus, _)] = 1

    def __filter(self, ruleset:'Ruleset') ->Generator[Tile, None, None]:
        """return element names"""
        return (x for x in self.occurrence
                if ruleset.withBonusTiles or (not x.isBonus))

    def count(self, ruleset:'Ruleset') ->int:
        """how many tiles are to be used by the game"""
        return cast(IntDict, self.occurrence).count(self.__filter(ruleset))

    def all(self, ruleset:'Ruleset') ->List[Tile]:
        """a list of all elements, each of them occurrence times"""
        return cast(IntDict, self.occurrence).all(self.__filter(ruleset))


class Piece(Tile):

    """
    This tile is part of the game. The wall is built from this.
    """

    def __new__(cls, *args: Any) ->'Piece':
        result = cast('Piece', cls._build(*args))
        return result

    def __init__(self, *args: Any) ->None:  # pylint: disable=unused-argument # type: ignore
        self.img:Optional['UITile'] = None

    def __hash__(self) ->int:
        """this is not inherited from Tile. I am sure there is a good reason."""
        return self.key


class PieceList(TileList):

    """a list that can only hold tiles"""

    tileClass = Piece

    def __contains__(self, value : Any) ->bool:
        """If value is Piece: must be the same object
        If value is Tile: any element having the same name"""
        if value.__class__ is Tile:
            _ =  value.name2()
            return any(x.name2() == _ for x in self)
        return any(x is value for x in self)

    def index(self, value : Tile, start: int =None , stop: int =None) ->int:  # type: ignore
        """Also accept Tile."""
        # FIXME: overload?
        if value.__class__ is Tile:
            for result, _ in enumerate(self):
                if _ == value:
                    return result
            raise ValueError(f'{value!r} is not in list {self!r}')
        return super().index(cast(Any, value), start,  stop)

    def remove(self, value : Tile) ->None:
        """Can also remove Tile."""
# FIXME: should we do tile == piece? would remove then work?
        if value.__class__ is Tile:
            name2 = value.name2()
            for _ in self:
                if _.name2() == name2:
                    value = _
                    break
            else:
                raise ValueError(f'{self} does not contain {value!r}')
        super().remove(cast(Any, value))

# those two must come first

Tile.unknownStr = 'Xy'
Tile.unknown = Tile(Tile.unknownStr)
Tile.noneStr = 'Xx'
Tile.none = Tile(Tile.noneStr)

elements = Elements()
assert not Tile.unknown.isKnown
for wind in Wind.all4:
    wind.tile = Tile('w', wind.char.lower())


class Meld(TileTuple, ReprMixin):

    """represents a meld. Can be empty. Many Meld methods will
    raise exceptions if the meld is empty. But we do not care,
    those methods are not supposed to be called on empty melds.
    Meld is essentially a list of Tile with added methods.
    A Meld is immutable, not from the view of python but for
    its user

    for melds with 3 tiles::
        isDeclared == isExposed : 3 exposed tiles
        not isDeclared == isConcealed: 3 concealed Tiles
        exposed = aaa
        exposedClaimed = aaa

    for melds with 4 tiles::
        isKong = aAAa or aaaa or aaaA but NOT AAAA
        isDeclared = aAAa or aaaa or aaaA
        isExposed = aaaa or aaaA
        isConcealed = aAAa or AAAA
        exposedClaimed = aaaA
        exposed = aaaa

    """
    # pylint: disable=too-many-instance-attributes

    cache : Dict[Any, 'Meld'] = {}

    def __new__(cls, iterable:Any=None) ->'Meld':
        """try to use cache"""
        if isinstance(iterable, str) and iterable in cls.cache:
            return cls.cache[iterable]
        if isinstance(iterable, Meld):
            # brauchen wir das?
            return iterable
        superclass:Type[Tiles] = TileTuple
        if hasattr(iterable, '__iter__') and not isinstance(iterable, str):
            iterable = list(iterable)
            if iterable:
                class1 = iterable[0].__class__
# FIXME: computeLastMelds may mix Piece and Tile, is that wanted?
                assert all(x.__class__ is class1 for x in iterable), repr(list(iterable))
                if class1 is Piece:
                    superclass = TileList
        tiles = cast(Meld, super(Meld, cls).__new__(cls, superclass(iterable)))
        if tiles in cls.cache:
            return cls.cache[tiles]
        return tiles

    def __init__(self, iterable:Any=None) ->None:
        """init the meld: content can be either
        - a single string with 2 chars for every tile
        - a list containing such strings
        - another meld. Its tiles are not passed.
        - a list of Tile objects"""
        # pylint:disable=too-many-statements
        self.exposed:'Meld'
        self.exposedClaimed:'Meld'
        self.declared:'Meld'
        self.concealed:'Meld'
        self.isPungKong:bool
        self.isPair:bool
        self.isPung:bool
        self.isKong:bool
        self.isDragonMeld:bool
        self.isHonorMeld:bool
        self.isWindMeld:bool
        if not hasattr(self, '_fixed'):  # already defined if I am from cache
            super().__init__(iterable)
            self.case = ''.join('a' if x.isExposed else 'A' for x in self)
            if self not in self.cache:
                self.cache[self] = self
                self.cache[str(self)] = self
            self.isExposed = self.__isExposed()
            self.isConcealed = not self.isExposed
            self.isSingle = self.isPair = self.isChow = self.isPung = False
            self.isKong = self.isClaimedKong = self.isKnitted = False
            self.isDragonMeld = len(self) > 0 and self[0].isDragon
            self.isWindMeld = len(self) > 0 and self[0].isWind
            self.isHonorMeld = self.isDragonMeld or self.isWindMeld
            self.isBonus = len(self) == 1 and self[0].isBonus
            self.isKnown = len(self) and self[0].isKnown
            self.__setMeldType()
            self.isPungKong = self.isPung or self.isKong
            self.isDeclared = self.isExposed or self.isKong
            groups = {x.group.lower() for x in self}
            if len(groups) == 1:
                self.group = self[0].group
                self.lowerGroup = self.group.lower()
            else:
                self.group = 'X'
                self.lowerGroup = 'x'
            self.isRest = False
            self.__staticRules:Dict[int, List['Rule']] = {}  # ruleset is key
            self.__dynamicRules:Dict[int, List['Rule']] = {}  # ruleset is key
            self.__staticDoublingRules:Dict[int, List['Rule']] = {}  # ruleset is key
            self.__dynamicDoublingRules:Dict[int, List['Rule']] = {}  # ruleset is key
            self.__hasRules:Optional[bool] = None
            self._fixed = True

            if len(self) < 4:
                TileTuple.__setattr__(
                    self,
                    'concealed',
                    Meld(x.concealed for x in self))
                super().__setattr__('declared', self.concealed)
                super().__setattr__('exposed', Meld(x.exposed for x in self))
                super().__setattr__('exposedClaimed', self.exposed)
            else:
                super().__setattr__('concealed', Meld(x.concealed for x in self))
                super().__setattr__('declared',
                    Meld([self[0].exposed, self[1].concealed, self[2].concealed, self[3].exposed]))
                super().__setattr__('exposed', Meld(x.exposed for x in self))
                super().__setattr__('exposedClaimed',
                    Meld([self[0].exposed, self[1].exposed, self[2].exposed, self[3].concealed]))

    def __setattr__(self, name:str, value:'Meld') ->None:
        if (hasattr(self, '_fixed')
                and not name.endswith('__hasRules')):
            raise TypeError
        super().__setattr__(name, value)

    def __prepareRules(self, ruleset:'Ruleset') ->None:
        """prepare rules from ruleset"""
        rulesetId = id(ruleset)
        self.__staticRules[rulesetId] = [
            x for x in ruleset.meldRules
            if not hasattr(x, 'mayApplyToMeld') and x.appliesToMeld(None, self)]
        self.__dynamicRules[rulesetId] = [
            x for x in ruleset.meldRules
            if hasattr(x, 'mayApplyToMeld') and x.mayApplyToMeld(self)]
        self.__hasRules = any(len(x) for x in chain(
            self.__staticRules.values(), self.__dynamicRules.values()))

        self.__staticDoublingRules[rulesetId] = [
            x for x in ruleset.doublingMeldRules
            if not hasattr(x, 'mayApplyToMeld') and hasattr(x, 'appliesToMeld') and x.appliesToMeld(None, self)]
        self.__dynamicDoublingRules[rulesetId] = [x for x in ruleset.doublingMeldRules
                                                  if hasattr(x, 'mayApplyToMeld') and x.mayApplyToMeld(self)]

    def rules(self, hand:'Hand') ->List['Rule']:
        """all applicable rules for this meld being part of hand"""
        if self.__hasRules is False:
            return []
        ruleset = hand.ruleset
        rulesetId = id(ruleset)
        if rulesetId not in self.__staticRules:
            self.__prepareRules(ruleset)
        result = self.__staticRules[rulesetId][:]
        result.extend(x for x in self.__dynamicRules[
            rulesetId] if x.appliesToMeld(hand, self))
        return result

    def doublingRules(self, hand:'Hand') ->List['Rule']:
        """all applicable doubling rules for this meld being part of hand"""
        ruleset = hand.ruleset
        rulesetId = id(ruleset)
        if rulesetId not in self.__staticRules:
            self.__prepareRules(ruleset)
        result = self.__staticDoublingRules[rulesetId][:]
        result.extend(x for x in self.__dynamicDoublingRules[
            rulesetId] if x.appliesToMeld(hand, self))
        return result

    def without(self, remove:Tile) ->TileTuple:
        """self without tile. The rest will be uppercased."""
        assert remove is not None, 'without(None) is illegal'
        _ = remove
        tiles = []
        for tile in self:
            if tile is _:
                _ = None
            else:
                tiles.append(tile.concealed)
        assert _ is None, f'trying to remove {_} from {self}'
        return TileTuple(tiles)

    def __setitem__(self, index:int, value:Tile) ->None:
        """set a tile in the meld"""
        raise TypeError

    def __delitem__(self, index:int) ->None:
        """removes a tile from the meld"""
        raise TypeError("'Meld' object doesn't support item deletion")

    def __isExposed(self) ->bool:
        """meld state: exposed or not"""
        if self.case.islower():
            return True
        if len(self) == 4:
            return self.case[1:3].islower()
        return False

    def __setMeldType(self) ->None:
        """compute meld type. Except knitting melds."""
        # pylint: disable=too-many-branches,too-many-return-statements
        length = len(self)
        if length == 0:
            return
        if length > 4:
            raise UserWarning(f'Meld {str(self)} is too long')
        if any(not x.isKnown for x in self):
            if len(set(self)) != 1:
                raise UserWarning(
                    'Meld %s: Cannot mix known and unknown tiles')
            self.isKnown = False
            return
        if length == 1:
            self.isSingle = True
            return
        if length == 2:
            if self[0] == self[1]:
                self.isPair = True
            elif self[0].value == self[1].value and self.case[0] == self.case[1] \
                    and all(x.lowerGroup in Tile.colors for x in self):
                self.isKnitted = True
            else:
                raise UserWarning(f'Meld {str(self)} is malformed')
            return
        # now length is 3 or 4
        tiles = set(self)
        if len(tiles) == 1:
            if length == 3:
                self.isPung = True
            elif self.case != 'AAAA':
                self.isKong = True
            return
        if len(tiles) == 3 and length == 3:
            if len({x.value for x in tiles}) == 1:
                if self.case in ('aaa', 'AAA'):
                    if len({x.group for x in tiles}) == 3:
                        if all(x.lowerGroup in Tile.colors for x in tiles):
                            self.isKnitted = True
                            return
        groups = {x.group for x in self}
        if len(groups) > 2 or len({x.lower() for x in groups}) > 1:
            raise UserWarning(f'Meld {str(self)} is malformed')
        values = {x.value for x in self}
        if length == 4:
            if len(values) > 1:
                raise UserWarning(f'Meld {str(self)} is malformed')
            if self.case == 'aaaA':
                self.isKong = self.isClaimedKong = True
            elif self.case == 'aAAa':
                self.isKong = True
            else:
                raise UserWarning(f'Meld {str(self)} is malformed')
            return
        # only possibilities left are CHOW and REST
        # length is 3
        if len(groups) == 1:
            if groups.pop().lower() in Tile.colors:
                if self[0].nextForChow is self[1] and self[1].nextForChow is self[2]:
                    self.isChow = True
                    return
        raise UserWarning(f'Meld {str(self)} is malformed')

    def __lt__(self, other:Tuple[Any, ...]) ->bool:
        """used for sorting. Smaller value is shown first."""
        if not other:
            return False
        if not self:
            return True
        if not isinstance(other, Meld):
            raise TypeError(f"'<' not supported between instances of {type(self).__name__} and {type(other).__name__}")
        if self.isDeclared and not cast(Meld, other).isDeclared:
            return True
        if not self.isDeclared and cast(Meld, other).isDeclared:
            return False
        if self[0].key == other[0].key:
            return len(self) > len(other)
        return self[0].key < other[0].key

    def typeName(self) ->str:
        """convert int to speaking name with shortcut. ATTENTION: UNTRANSLATED!"""
        # pylint: disable=too-many-return-statements
        if self.isBonus:
            return i18nc('kajongg meld type', 'Bonus')
        if self.isSingle:
            return i18nc('kajongg meld type', '&single')
        if self.isPair:
            return i18nc('kajongg meld type', '&pair')
        if self.isChow:
            return i18nc('kajongg meld type', '&chow')
        if self.isPung:
            return i18nc('kajongg meld type', 'p&ung')
        if self.isClaimedKong:
            return i18nc('kajongg meld type', 'c&laimed kong')
        if self.isKong:
            return i18nc('kajongg meld type', 'k&ong')
        return i18nc('kajongg meld type', 'rest of tiles')

    def __stateName(self) ->str:
        """the translated name of the state"""
        if self.isBonus or self.isClaimedKong:
            return ''
        if self.isExposed:
            return i18nc('kajongg meld state', 'Exposed')
        return i18nc('kajongg meld state', 'Concealed')

    def name(self) ->str:
        """the long name"""
        result = i18nc(
            'kajongg meld name, do not translate parameter names',
            '{state} {meldType} {name}')
        return result.format(
            state=self.__stateName(),
            meldType=self.typeName(),
            name=self[0].name()).replace('  ', ' ').strip()

    @staticmethod
    def cacheMeldsInTiles() ->None:
        """define all usual melds as Tile attributes"""
        Tile.unknown.single = Meld(Tile.unknown)
        Tile.unknown.pung = Meld(Tile.unknown * 3)
        do_those = list(elements.occurrence.keys())
        for tile in do_those:
            tile.cacheMelds()

    def __repr__(self):
        """because TileTuple.__repr__ does not indicate class Meld"""
        return f'Meld({TileTuple.__repr__(self)}'


class MeldList(list):

    """a list of melds"""

    def __init__(self, newContent:Any=None) ->None:
        list.__init__(self)
        if newContent is None:
            return
        if isinstance(newContent, Meld):
            list.append(self, newContent)
        elif isinstance(newContent, (list, GeneratorType)):
            # I tried hasattr('__iter__') but that does not work
            list.extend(self, newContent)
        elif isinstance(newContent, str):
            list.extend(self, [Meld(x)
                               for x in newContent.split()])
        else:
            list.extend(self, [Meld(x) for x in newContent])
        self.sort()

    def __getitem__(self, key):
        if isinstance(key, slice):
            return type(self)(list.__getitem__(self, key))
        return super().__getitem__(key)

    def extend(self, values:Iterable[Meld]) ->None:
        super().extend(values)
        self.sort()

    def append(self, value:Meld) ->None:
        super().append(value)
        self.sort()

    def tiles(self) ->TileTuple:
        """flat view of all tiles in all melds"""
        return TileTuple(self)

    def __str__(self) ->str:
        if self:
            return ' '.join(str(x) for x in self)
        return ''

Meld.cacheMeldsInTiles()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only


The DBPasswordChecker is based on an example from the book
Twisted Network Programming Essentials by Abe Fettig, 2006
O'Reilly Media, Inc., ISBN 0-596-10032-9
"""

# pylint: disable=wrong-import-order, wrong-import-position

import sys
import os
import logging
import datetime
import argparse
from typing import TYPE_CHECKING, Tuple, Any, Optional, Sequence, List, Mapping, Dict, Union, cast, Type

from zope.interface import implementer


def cleanExit(*unusedArgs: Any) ->None:
    """we want to cleanly close sqlite3 files"""
    if Debug.quit:
        logDebug('cleanExit')
    if Options.socket and sys.platform != 'win32':
        try:
            os.remove(Options.socket)
        except FileNotFoundError:
            pass
    try:
        if Internal.db:
            Internal.db.close()
                              # setting to None does not call close(), do we
                              # need close?
        logging.shutdown()
        os._exit(0)
    except NameError:
        logging.shutdown()
    try:
        reactor.stop()  # type:ignore[misc]
    except NameError:
        sys.exit(0)
    except ReactorNotRunning:
        pass

from common import handleSignals
handleSignals(cleanExit)

from common import Options, Internal, Debug
Internal.isServer = True
Internal.logPrefix = 'S'

from twisted.spread import pb
from twisted.internet.interfaces import IReactorCore
from twisted.internet import error
from twisted.internet.defer import maybeDeferred, fail, succeed, Deferred
from twisted.cred import checkers, portal, credentials, error as credError
from twisted.internet import reactor as reactor_module
from twisted.internet.error import ReactorNotRunning
if TYPE_CHECKING:
    from twisted.python.failure import Failure

reactor = cast(IReactorCore, reactor_module)
reactor.addSystemEventTrigger('before', 'shutdown', cleanExit)  # type:ignore[arg-type]
Internal.reactor = reactor  # type:ignore[assignment]

from player import Players
from query import Query, initDb
from log import logDebug, logWarning, logError, logInfo, logException, SERVERMARK, logFailure
from mi18n import i18n, i18nE
from util import elapsedSince
from message import Message, ChatMessage
from deferredutil import DeferredBlock
from rule import Ruleset
from servercommon import srvError, srvMessage
from user import User
from servertable import ServerTable, ServerGame


@implementer(checkers.ICredentialsChecker)
class DBPasswordChecker:

    """checks against our sqlite3 databases"""
    credentialInterfaces = (credentials.IUsernamePassword,
                            credentials.IUsernameHashedPassword)

    def requestAvatarId(self, cred: pb._PortalAuthChallenger) ->Deferred:
        """get user id from database"""
        cred.username = cred.username.decode('utf-8')
        args = cred.username.split(SERVERMARK)
        if len(args) > 1:
            if args[0] == 'adduser':
                cred.username = args[1]
                password = args[2]
                query = Query(
                    'insert or ignore into player(name,password) values(?,?)',
                    (cred.username,
                     password))
            elif args[1] == 'deluser':
                pass
        query = Query(
            'select id, password from player where name=?', (cred.username,))
        if not query.records:
            template = 'Wrong username: %1'
            if Debug.connections:
                logDebug(i18n(template, cred.username))
            return fail(credError.UnauthorizedLogin(srvMessage(template, cred.username)))
        userid, password = query.records[0]
        defer1 = maybeDeferred(cred.checkPassword, password.encode('utf-8'))
        defer1.addCallback(DBPasswordChecker._checkedPassword, userid).addErrback(logFailure)
        return defer1

    @staticmethod
    def _checkedPassword(matched: bool, userid:str) ->Union[str, Deferred]:
        """after the password has been checked. TODO: is the result type correct?"""
        if not matched:
            return fail(credError.UnauthorizedLogin(srvMessage(i18nE('Wrong password'))))
        return userid


class MJServer:

    """the real mah jongg server"""

    def __init__(self) ->None:
        self.tables:Dict[int, ServerTable] = {}
        self.srvUsers:List[User] = []
        Players.load()
        self.lastPing = datetime.datetime.now()
        self.checkPings()

    def chat(self, chatString:str) ->None:
        """a client sent us a chat message"""
        # FIXME: why does this work?
        chatLine = ChatMessage(chatString)  # type:ignore[arg-type]


        self.tables[chatLine.tableid].sendChatMessage(chatLine)

    def login(self, user: User) ->None:
        """accept a new user"""
        if user not in self.srvUsers:
            self.srvUsers.append(user)
            self.loadSuspendedTables(user)

    def callRemote(self, user:User, *args: Any, **kwargs:Mapping[Any, Any]) ->Deferred:
        """if we still have a connection, call remote, otherwise clean up"""
        if user.mind:
            try:
                args2, kwargs2 = Message.jellyAll(args, kwargs)
                return user.mind.callRemote(*args2, **kwargs2).addErrback(MJServer.ignoreLostConnection)
            except (pb.DeadReferenceError, pb.PBConnectionLost):
                user.mind = None
                self.logout(user)
        return succeed([])

    @staticmethod
    def __stopAfterLastDisconnect() ->None:
        """as the name says"""
        if Options.socket and not Options.continueServer:
            try:
                reactor.stop()  # type:ignore[misc]
                if Debug.connections:
                    logDebug(f'local server terminates from {Options.socket}. Reason: last client disconnected')
            except ReactorNotRunning:
                pass

    def checkPings(self) ->None:
        """are all clients still alive? If not log them out"""
        since = elapsedSince(self.lastPing)
        if self.srvUsers and since > 30:
            if Debug.quit:
                logDebug(f'no ping since {elapsedSince(self.lastPing)} seconds but we still have users:{self.srvUsers}')
        if not self.srvUsers and since > 30:
            # no user at all since 30 seconds, but we did already have a user
            self.__stopAfterLastDisconnect()
        for user in self.srvUsers:
            if elapsedSince(user.lastPing) > 60:
                logInfo(
                    f'No messages from {user.name} since 60 seconds, clearing connection now')
                user.mind = None
                self.logout(user)
        reactor.callLater(10, self.checkPings)

    @staticmethod
    def ignoreLostConnection(failure: 'Failure') ->None:
        """if the client went away correctly, do not dump error messages on stdout."""
        msg = failure.getErrorMessage()
        if 'twisted.internet.error.ConnectionDone' not in msg:
            logError(msg)
        failure.trap(pb.PBConnectionLost)

    def sendTables(self, user:User, tables:Optional[List[ServerTable]]=None) ->Deferred:
        """send tables to user. If tables is None, he gets all new tables and those
        suspended tables he was sitting on"""
        if tables is None:
            tables = [
                x for x in self.tables.values()
                if not x.running and (not x.suspendedAt or x.hasName(user.name))]
        if tables:
            data = [x.asSimpleList() for x in tables]
            if Debug.table:
                logDebug(
                    f'sending {len(tables)} tables to {user.name}: {data}')
            return self.callRemote(user, 'newTables', data)
        return succeed([])

    def _lookupTable(self, tableid: int) ->ServerTable:
        """return table by id or raise exception"""
        if tableid not in self.tables:
            srvError(
                pb.Error,
                i18nE('table with id <numid>%1</numid> not found'),
                tableid)
        return self.tables[tableid]

    def generateTableId(self) ->int:
        """generates a new table id: the first free one"""
        usedIds = set(self.tables or [0])
        availableIds = set(x for x in range(1, 2 + max(usedIds)))
        return min(availableIds - usedIds)

    def newTable(self, user:User, ruleset:str, playOpen:bool,
                 autoPlay:bool, wantedGame:str, tableId:Optional[int]=None) ->Optional[Deferred]:
        """user creates new table and joins it"""
        def gotRuleset(ruleset:str) ->None:
            """now we have the full ruleset definition from the client"""
            Ruleset.cached(
                ruleset).save()  # make it known to the cache and save in db
        if tableId in self.tables:
            assert tableId is not None  # for mypy
            return fail(srvError(pb.Error,
                                 f'You want a new table with id={int(tableId)} '
                                 f'but that id is already used for table {self.tables[tableId]}'))
        if Ruleset.hashIsKnown(ruleset):
            return self.__newTable(None, user, ruleset, playOpen, autoPlay, wantedGame, tableId)
        _ = self.callRemote(user, 'needRuleset', ruleset)
        if _:
            return _.addCallback(
                gotRuleset).addErrback(logException).addCallback(
                    self.__newTable, user, ruleset, playOpen, autoPlay, wantedGame, tableId).addErrback(logException)
        return None

    def __newTable(self, unused:None, user:User, ruleset:str,
                   playOpen:bool, autoPlay:bool, wantedGame:str, tableId:Optional[int]=None) ->Deferred:
        """now we know the ruleset"""
        def sent(unused: Any) ->int:
            """new table sent to user who created it"""
            return table.tableid
        table = ServerTable(
            self,
            owner=user,
            ruleset=ruleset,
            suspendedAt=None,
            playOpen=playOpen,
            autoPlay=autoPlay,
            wantedGame=wantedGame,
            tableId=tableId)
        result = None
        for srvUser in self.srvUsers:
            deferred = self.sendTables(srvUser, [table])
            if user == srvUser:
                result = deferred
                deferred.addCallback(sent).addErrback(logFailure)
        assert result
        return result

    def needRulesets(self, rulesetHashes: List[str]) -> List[List[List[Union[int, str, float]]]]:
        """the client wants those full rulesets"""
        result = []
        for table in self.tables.values():
            if table.ruleset.hash in rulesetHashes:
                result.append(table.ruleset.toList())
        return result

    def joinTable(self, user: User, tableid: int) ->bool:
        """user joins table"""
        table = self._lookupTable(tableid)
        table.addUser(user)
        block = DeferredBlock(table, where='joinTable')
        block.tell(
            None,
            self.srvUsers,
            Message.TableChanged,
            source=table.asSimpleList())
        if len(table.users) == table.maxSeats():
            if Debug.table:
                logDebug(f'Table {table}: All seats taken, starting')

            def startTable(unused: Any) -> None:
                """now all players know about our join"""
                assert table.owner
                table.readyForGameStart(table.owner)
            block.callback(startTable)
        else:
            block.callback(False)
        return True

    def tablesWith(self, user: User) ->List[int]:
        """table ids with user, except table 'without'"""
        return [x.tableid for x in self.tables.values() if user in x.users]

    def leaveTable(self, user: User, tableid: int, message: str, *args: str) ->bool:
        """user leaves table. If no human user is left on a new table, remove it"""
        if tableid in self.tables:
            table = self.tables[tableid]
            if user in table.users:
                if len(table.users) == 1 and not table.suspendedAt:
                    # silent: do not tell the user who left the table that he
                    # did
                    self.removeTable(table, 'silent', message, *args)
                else:
                    table.delUser(user)
                    if self.srvUsers:
                        block = DeferredBlock(table, where='leaveTable')
                        block.tell(
                            None,
                            self.srvUsers,
                            Message.TableChanged,
                            source=table.asSimpleList())
                        block.callback(False)
        return True

    def startGame(self, user: User, tableid: int) ->None:
        """try to start the game"""
        self._lookupTable(tableid).readyForGameStart(user)

    def removeTable(self, table: ServerTable, reason: str, message: str, *args:Union[str, int]) ->None:
        """remove a table"""
        assert reason in ('silent', 'tableRemoved', 'gameOver', 'abort')
        # HumanClient implements methods remote_tableRemoved etc.
        if Debug.connections or reason == 'abort':
            logDebug(
                f"{f'{int(table.game.seed)}:' if table.game else ''}{i18n(message, *args)} ", withGamePrefix=False)
        if table.tableid in self.tables:
            del self.tables[table.tableid]
            if reason == 'silent':
                tellUsers = []
            else:
                tellUsers = table.users if table.running else self.srvUsers
            for user in tellUsers:
                # this may in turn call removeTable again!
                self.callRemote(user, reason, table.tableid, message, *args)
            for user in table.users:
                table.delUser(user)
            if Debug.table:
                logDebug(
                    f'removing table {int(table.tableid)}: {i18n(message, *args)} {reason}')
        if table.game:
            table.game.close()

    def logout(self, user: User) ->None:
        """remove user from all tables"""
        if user not in self.srvUsers:
            return
        self.srvUsers.remove(user)
        for tableid in self.tablesWith(user):
            self.leaveTable(
                user,
                tableid,
                i18nE('Player %1 has logged out'),
                user.name)
        # wait a moment. We want the leaveTable message to arrive everywhere before
        # we say serverDisconnects. Sometimes the order was reversed.
        reactor.callLater(1, self.__logout2, user)

    def __logout2(self, user: User) ->None:
        """now the leaveTable message had a good chance to get to the clients first"""
        self.callRemote(user, 'serverDisconnects')
        user.mind = None
        for block in DeferredBlock.blocks:
            for request in block.requests:
                if request.user == user:
                    request.answer = Message.Abort

    def loadSuspendedTables(self, user: User) ->None:
        """loads all yet unloaded suspended tables where this
        user is participating. We do not unload them if the
        user logs out, there are filters anyway returning only
        the suspended games for a certain user.
        Never load old autoplay games."""
        query = Query("select distinct g.id, g.starttime, "
                      "g.seed, "
                      "ruleset, s.scoretime "
                      "from game g, player p0, score s,"
                      "player p1, player p2, player p3 "
                      "where autoplay=0 "
                      " and p0.id=g.p0 and p1.id=g.p1 "
                      " and p2.id=g.p2 and p3.id=g.p3 "
                      " and (p0.name=? or p1.name=? or p2.name=? or p3.name=?) "
                      " and s.game=g.id"
                      " and g.endtime is null"
                      " and exists(select 1 from ruleset where ruleset.id=g.ruleset)"
                      " and exists(select 1 from score where game=g.id)"
                      " and s.scoretime = (select max(scoretime) from score where game=g.id) limit 10",
                      (user.name, user.name, user.name, user.name))
        for gameid, _, seed, ruleset, suspendTime in query.records:
            if gameid not in (x.game.gameid for x in self.tables.values() if x.game):
                table = ServerTable(
                    self, None, ruleset, suspendTime, playOpen=False,
                    autoPlay=False, wantedGame=str(seed))
                table.game = cast(ServerGame, ServerGame.loadFromDB(gameid))


@implementer(portal.IRealm)
class MJRealm:

    """connects mind and server"""

    def __init__(self) ->None:
        self.server : Optional[MJServer] = None

    def requestAvatar(self, avatarId: str, mind: pb.RemoteReference,
        *interfaces: Sequence[pb.IPerspective]) -> Tuple[Type[pb.IPerspective], User, Any]:
        """as the tutorials do..."""
        if pb.IPerspective not in interfaces:
            raise NotImplementedError("No supported avatar interface")
        avatar = User(avatarId)
        avatar.server = self.server
        avatar.attached(mind)
        if Debug.connections:
            logDebug(f'Connection from {avatar.source()} ')
        return pb.IPerspective, avatar, lambda a=avatar: a.detached(mind)

def parseArgs() -> argparse.Namespace:
    """as the name says"""
    parser = argparse.ArgumentParser()
    defaultPort = Internal.defaultPort
    parser.add_argument('--port', dest='port',
                      help=i18n(
                          f'the server will listen on PORT ({int(defaultPort)})'),
                      type=int, default=defaultPort)
    parser.add_argument('--socket', dest='socket',
                      help=i18n('the server will listen on SOCKET'), default=None)
    parser.add_argument(
        '--db',
        dest='dbpath',
        help=i18n('name of the database'),
        default=None)
    parser.add_argument(
        '--continue', dest='continueServer', action='store_true',
        help=i18n('do not terminate local game server after last client disconnects'), default=False)
    parser.add_argument('--debug', dest='debug',
                      help=Debug.help())
    args = parser.parse_args(sys.argv[1:])
    Options.continueServer |= args.continueServer
    if args.dbpath:
        Options.dbPath = os.path.expanduser(args.dbpath)
    if args.socket:
        Options.socket = args.socket
    Debug.setOptions(args.debug)
    Options.fixed = True  # may not be changed anymore
    del parser           # makes Debug.gc quieter
    return args


def kajonggServer() ->None:
    """start the server"""
    options = parseArgs()
    if not initDb():
        sys.exit(1)
    realm = MJRealm()
    realm.server = MJServer()
    kajonggPortal = portal.Portal(realm, [DBPasswordChecker()])  # type: ignore[arg-type,list-item]
    import predefined
    predefined.load()
    try:
        if Options.socket:
            # we do not want tracebacks to go from server to client,
            # please check on the server side instead
            factory = pb.PBServerFactory(kajonggPortal, unsafeTracebacks=False)
            if sys.platform == 'win32':
                if Debug.connections:
                    logDebug(
                        f'local server listening on 127.0.0.1 port {int(options.port)}')
                reactor.listenTCP(options.port, factory, interface='127.0.0.1')
            else:
                if Debug.connections:
                    logDebug(
                        f'local server listening on UNIX socket {Options.socket}')
                reactor.listenUNIX(Options.socket, factory)
        else:
            if Debug.connections:
                logDebug(f'server listening on port {int(options.port)}')
            reactor.listenTCP(options.port, pb.PBServerFactory(kajonggPortal))
    except error.CannotListenError as errObj:
        logWarning(errObj)
        sys.exit(1)
    else:
        reactor.run()  # type:ignore[misc]


def profileMe() ->None:
    """where do we lose time?"""
    import cProfile
    cProfile.run('kajonggServer()', 'prof')
    import pstats
    statistics = pstats.Stats('prof')
    statistics.sort_stats('cumulative')
    statistics.print_stats(40)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# pylint: disable=wrong-import-position

import sys
import os
import logging
from typing import Tuple, List, Optional, Type, Any

from qt import QObject, QCommandLineParser, QCommandLineOption, Qt, QGuiApplication
from kde import KApplication
from mi18n import i18n, MLocale
from point import PointRange

from common import Options, SingleshotOptions, Internal, Debug
# do not import modules using twisted before our reactor is running

def __initRulesetsOrExit() ->None:
    """exits if user only wanted to see available rulesets
    or if the given ruleset does not exist"""
    import predefined
    predefined.load()
    if Options.showRulesets or Options.rulesetName:
        from rule import Ruleset
        rulesets = {x.name: x for x in Ruleset.selectableRulesets()}
        if Options.showRulesets:
            for name in rulesets:
                print(name)
            Internal.db.close()
            sys.exit(0)
        elif Options.rulesetName in rulesets:
            # we have an exact match
            Options.ruleset = rulesets[Options.rulesetName]
        else:
            matches = [x for x in rulesets if Options.rulesetName in x]
            if len(matches) != 1:
                if not matches:
                    msg = f'Ruleset {Options.rulesetName} is unknown'
                else:
                    msg = f"Ruleset {Options.rulesetName} is ambiguous: {', '.join(matches)}"
                Internal.db.close()
                raise SystemExit(msg)
            Options.ruleset = rulesets[matches[0]]


class CommandLineOption(QCommandLineOption):
    """add some helping attributes"""
    def __init__(self, name : str, description : str, valueName : Optional[str] =None,
        defaultValue : Optional[str]=None, optName : Optional[str]=None,
        argType : Optional[Type]=None, singleshot:bool=False) ->None:
        super().__init__([name], description, valueName or '', defaultValue or '')
        if argType is None:
            if valueName is None:
                argType = bool
            else:
                argType = str
        self.argType = argType
        self.optName = optName or name
        self.singleshot = singleshot

def defineOptions() -> Tuple[QCommandLineParser, List[CommandLineOption]]:
    """define command line options"""
    parser = QCommandLineParser()
    options = []
    def option(name:str, description:str, valueName:Optional[str]=None, defaultValue:Optional[str]=None,
        optName:Optional[str]=None, argType:Optional[Type]=None, singleshot:bool=False) ->None:
        """helper"""
        opt = CommandLineOption(name, description, valueName, defaultValue,
            optName=optName, argType=argType, singleshot=singleshot)
        options.append(opt)
        parser.addOption(opt)

    parser.setApplicationDescription(i18n('Mah Jongg - the ancient Chinese board game for 4 players'))

    parser.addHelpOption()
    parser.addVersionOption()
    option('playopen', i18n('all robots play with visible concealed tiles'), optName='playOpen')
    option('demo', i18n('start with demo mode'))
    option('host', i18n("login to HOST"), 'HOST', '')
    option('table', i18n('start new TABLE'), 'TABLE', '1', argType=int, singleshot=True)
    option('join', i18n('join TABLE'), 'TABLE', '1', argType=int, singleshot=True)
    option('ruleset', i18n('use RULESET without asking'), 'RULESET', '', optName='rulesetName')
    option('player', i18n('prefer PLAYER for next login'), 'PLAYER', '')
    option('ai', i18n('use AI variant for human player in demo mode'), 'AI', '', optName='AI')
    option('csv', i18n('write statistics to CSV'), 'CSV', '')
    option('rulesets', i18n('show all available rulesets'), optName='showRulesets')
    option('game', i18n('for testing purposes: Initializes the random generator'),
           'seed(/firsthand)(..(lasthand))', '0')
    option('nogui', i18n('show no graphical user interface. Intended only for testing'), optName='gui')
    option('socket', i18n(
        'use a dedicated server already running and listening on SOCKET. Intended only for testing'), 'SOCKET', '')
    option('port', i18n(
        'use a dedicated server already running and listening on PORT. Intended only for testing'), 'PORT', '')
    option('debug', Debug.help(), 'DEBUG', '')
    return parser, options

def parseOptions() ->None:
    """parse command line options and save the values"""
    Options.gui = True
    parser, options = defineOptions()
    parser.process(Internal.app)
    for option in options:
        if parser.isSet(option):
            value = parser.value(option) if option.valueName() else ''
            if option.optName == 'debug':
                msg = Debug.setOptions(value)
                if msg:
                    Internal.logger.debug(msg)
                    logging.shutdown()
                    sys.exit(2)
                continue
            target: Type
            if option.optName in SingleshotOptions.__dict__:
                target = SingleshotOptions
            else:
                target = Options
            if option.argType is bool:
                setattr(target, option.optName, not option.names()[0].startswith('no'))
            elif option.argType is int:
                setattr(target, option.optName, int(value))
            else:
                setattr(target, option.optName, value)

    Options.demo |= not Options.gui
    Internal.autoPlay = Options.demo
    if SingleshotOptions.game:
        PointRange.from_string(SingleshotOptions.game)  # checks for valid syntax

    from query import initDb
    if not initDb():
        raise SystemExit('Cannot initialize database')
    __initRulesetsOrExit()
    Options.fixed = True  # may not be changed anymore


class EvHandler(QObject):

    """an application wide event handler"""

    def eventFilter(self, receiver: Any, event: Any) ->bool:
        """will be called for all events"""
        from log import EventData
        EventData(receiver, event)
        return super().eventFilter(receiver, event)

from util import gitHead

if sys.platform == 'win32':
    _ = os.path.dirname(os.path.realpath(__file__))
    if _.endswith('.zip'):
        # cx_freeze
        os.chdir(os.path.dirname(_))

QGuiApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts, True)

Internal.app = KApplication()
parseOptions()

if hasattr(QGuiApplication, 'setDesktopFileName'):
    QGuiApplication.setDesktopFileName('org.kde.kajongg')

if Debug.neutral:
    MLocale.translation = None

if Debug.locate:
    # this has been read before Debug.locate is set
    Internal.logger.debug('Configuration in %s', Internal.kajonggrc.path)

if Debug.events:
    EVHANDLER = EvHandler()
    Internal.app.installEventFilter(EVHANDLER)

from config import SetupPreferences
SetupPreferences()

if Options.csv:
    if gitHead() == 'current':
        Internal.logger.debug(
            'You cannot write to %s with changes uncommitted to git',
            Options.csv)
        sys.exit(2)

from mainwindow import MainWindow
MainWindow()
Internal.app.exec()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# pylint: disable=invalid-name

import inspect
from typing import TYPE_CHECKING, Optional

from twisted.internet.defer import Deferred, succeed

from kde import KMessageBox, KDialog

from qt import Qt, QDialog, QMessageBox, QWidget

from common import Options, Internal, isAlive, ReprMixin

if TYPE_CHECKING:
    from qt import QEvent, QKeyEvent, QPushButton, QIcon, QDialogButtonBox

class KDialogIgnoringEscape(KDialog):

    """as the name says"""

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """catch and ignore the Escape key"""
        if event:
            if event.key() == Qt.Key.Key_Escape:
                event.ignore()
            else:
                super().keyPressEvent(event)


class MustChooseKDialog(KDialogIgnoringEscape):

    """this dialog can only be closed if a choice has been done. Currently,
    the self.chosen thing is not used, code removed.
    So this dialog can only be closed by calling accept() or reject()"""

    def __init__(self) ->None:
        parent:Optional[QWidget] = Internal.mainWindow  # default
        # if we are (maybe indirectly) called from a method belonging to a QWidget, take that as parent
        # this does probably not work for classmethod or staticmethod but it is
        # good enough right now
        for frametuple in inspect.getouterframes(inspect.currentframe())[1:]:
            if 'self' in frametuple[0].f_locals:
                obj = frametuple[0].f_locals['self']
                if isinstance(obj, QWidget) and not isinstance(obj, QDialog) and isAlive(obj):
                    parent = obj
                    break
        if not isAlive(parent):
            parent = None
        super().__init__(parent)

    def closeEvent(self, event:Optional['QEvent']) ->None:
        """self.chosen is currently not used, never allow this"""
        if event:
            event.ignore()


class Prompt(MustChooseKDialog, ReprMixin):

    """common code for things like QuestionYesNo, Information"""

    def __init__(self, msg:str, icon:'QMessageBox.Icon'=QMessageBox.Icon.Information,
                 buttons:'QDialogButtonBox.StandardButton'=KDialog.Ok,
                 caption:Optional[str]=None, default:Optional['QDialogButtonBox.StandardButton']=None) ->None:
        """buttons is button codes or-ed like KDialog.Ok | KDialog.Cancel. First one is default."""
        if r'\n' in msg:
            print(r'*********************** Fix this! Prompt gets \n in', msg)
            msg = msg.replace(r'\n', '\n')
        self.msg = msg
        self.default = default
        if Options.gui:
            super().__init__()
            self.setWindowFlags(self.windowFlags() | Qt.WindowType.WindowStaysOnTopHint)
            self.setCaption(caption or '')
            KMessageBox.createKMessageBox(
                self, icon, msg,
                [], "", False,
                KMessageBox.Options(KMessageBox.NoExec | KMessageBox.AllowLink))
            self.setButtons(buttons)
            # buttons is either Yes/No or Ok
            defaultButton = KDialog.Yes if KDialog.Yes & buttons else KDialog.Ok
            assert defaultButton & buttons, buttons
            self.button(defaultButton).setFocus()

    def __str__(self) ->str:
        return self.msg


class DeferredDialog(Deferred):

    """make dialogs usable as Deferred"""

    def __init__(self, dlg:KDialog, modal:bool=True, always:bool=False) ->None:
        super().__init__()
        self.dlg:Optional[KDialog] = dlg
        self.modal = modal
        self.always = always
        if Options.gui:
            if hasattr(self.dlg, 'buttonClicked'):
                self.dlg.buttonClicked.connect(self.clicked)
            else:
                self.dlg.accepted.connect(self.clicked)
                self.dlg.rejected.connect(self.cancel)
        if hasattr(Internal, 'reactor'):
            # sometimes pylint 2.7.2 warns, sometimes not
            Internal.reactor.callLater(0, self.__execute)
        else:
            # we do not yet have a reactor in initDb()
            self.__execute()

    def __execute(self) ->None:
        """now do the actual action"""
        if self.dlg is None:
            return
        scene = Internal.scene
        if not Options.gui or not isAlive(self.dlg):
            self.clicked()
            return
        autoPlay = bool(scene and scene.game and scene.game.autoPlay)
        autoAnswerDelayed = autoPlay and not self.always
        if self.modal and not autoAnswerDelayed:
            self.dlg.exec()
        else:
            self.dlg.show()
        if autoAnswerDelayed:
            assert Internal.Preferences
            Internal.reactor.callLater(
                Internal.Preferences.animationDuration() / 500.0,
                self.clicked)

    def clicked(self, button:Optional['QDialogButtonBox.StandardButton']=None) ->None:
        """we got a reaction"""
        if self.dlg:
            result = self.dlg.returns(button)
        else:
            result = None
        self.__removeFromScene()
        self.callback(result)

    def cancel(self) ->None:
        """we want no answer, just let the dialog disappear"""
        self.__removeFromScene()
        super().cancel()

    def __removeFromScene(self) ->None:
        """remove ourself"""
        if self.dlg and Internal.scene and isAlive(self.dlg):
            self.dlg.hide()
        self.dlg = None


class QuestionYesNo(DeferredDialog):

    """wrapper, see class Prompt"""

    def __init__(self, msg:str, modal:bool=True, always:bool=False, caption:Optional[str]=None) ->None:
        dialog = Prompt(msg, icon=QMessageBox.Icon.Question,
                        buttons=KDialog.Yes | KDialog.No, default=KDialog.Yes, caption=caption)
        super().__init__(dialog, modal=modal, always=always)


class WarningYesNo(DeferredDialog):

    """wrapper, see class Prompt"""

    def __init__(self, msg:str, modal:bool=True, caption:Optional[str]=None) ->None:
        dialog = Prompt(msg, icon=QMessageBox.Icon.Warning,
                        buttons=KDialog.Yes | KDialog.No, default=KDialog.Yes, caption=caption)
        super().__init__(dialog, modal=modal)


class Information(DeferredDialog):

    """wrapper, see class Prompt"""

    def __init__(self, msg:str, modal:bool=True, caption:Optional[str]=None) ->None:
        dialog = Prompt(msg, icon=QMessageBox.Icon.Information,
                        buttons=KDialog.Ok, caption=caption)
        super().__init__(dialog, modal=modal)


class Sorry(DeferredDialog):

    """wrapper, see class Prompt"""

    def __init__(self, msg:str, modal:bool=True, caption:Optional[str]=None, always:bool=False) ->None:
        dialog = Prompt(msg, icon=QMessageBox.Icon.Information,
                        buttons=KDialog.Ok, caption=caption or 'Sorry')
        super().__init__(dialog, modal=modal, always=always)


def NoPrompt(unusedMsg:str) ->Deferred:
    """we just want to be able to add callbacks even if non-interactive"""
    return succeed(None)



"""
Authors of original libkmahjongg in C++:
    Copyright (C) 1997 Mathias Mueller <in5y158@public.uni-hamburg.de>
    Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

this python code:
    Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import TYPE_CHECKING, Optional


from qt import Qt, QPainter, QBrush, QPalette, QPixmapCache, QPixmap
from qt import QSvgRenderer

from log import logException, i18n
from mjresource import Resource

if TYPE_CHECKING:
    from qt import QWidget, QSize

class Background(Resource):

    """represents a background"""

    resourceName = 'background'
    configGroupName = 'KMahjonggBackground'
    cache = {}

    def __init__(self, name:Optional[str]=None) ->None:
        """continue __build"""
        super().__init__(name)
        self.graphicsPath = None
        QPixmapCache.setCacheLimit(20480)  # the chinese landscape needs much

        self.tiled = self.group.readEntry('Tiled') == '1'
        if self.tiled:
            try:
                self.imageWidth = self.group.readInteger('Width')
                self.imageHeight = self.group.readInteger('Height')
            except Exception as exc:
                logException(exc)  # TODO: simplify if we switch to twisted logger
                raise
        self.isPlain = bool(self.group.readEntry('Plain'))
        if not self.isPlain:
            graphName = self.group.readEntry("FileName")
            assert isinstance(graphName, str)
            self.graphicsPath = self.locate(graphName)
            if not self.graphicsPath:
                logException(
                    f'cannot find kmahjongglib/backgrounds/{graphName} for {self.desktopFileName}')

    def __pixmap(self, size:'QSize') ->QPixmap:
        """return a background pixmap"""
        if self.isPlain:
            return QPixmap()
        width = size.width()
        height = size.height()
        if self.tiled:
            width = self.imageWidth
            height = self.imageHeight
        cachekey = f'{self.name}W{width}H{height}'
        result = QPixmapCache.find(cachekey)  # type:ignore[call-overload]
        if not result:
            renderer = QSvgRenderer(self.graphicsPath)
            if not renderer.isValid():
                logException(
                    i18n('file <filename>%1</filename> contains no valid SVG', self.graphicsPath))
            result = QPixmap(int(width), int(height))
            result.fill(Qt.GlobalColor.transparent)
            painter = QPainter(result)
            renderer.render(painter)
            QPixmapCache.insert(cachekey, result)
        return result

    def brush(self, size:'QSize') ->QBrush:
        """background brush"""
        return QBrush(self.__pixmap(size))

    def setPalette(self, onto:'QWidget') ->None:
        """set a background palette for widget onto"""
        palette = QPalette()
        mybrush = self.brush(onto.size())
        palette.setBrush(QPalette.ColorRole.Window, mybrush)
        onto.setPalette(palette)
